# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2023 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OSCAL transformation tasks."""

# mypy: ignore-errors  # noqa E800
import configparser
import csv
import datetime
import logging
import os
import pathlib
import sys
import traceback
import uuid
from math import log10
from typing import Generator, Iterator, List, Optional, Union

from trestle.common.list_utils import as_list
from trestle.core.catalog.catalog_interface import CatalogInterface
from trestle.core.profile_resolver import ProfileResolver
from trestle.oscal import OSCAL_VERSION
from trestle.oscal.common import Metadata
from trestle.oscal.common import Property
from trestle.oscal.component import ComponentDefinition
from trestle.oscal.component import ControlImplementation
from trestle.oscal.component import DefinedComponent
from trestle.oscal.component import ImplementedRequirement
from trestle.oscal.component import SetParameter
from trestle.oscal.component import Statement
from trestle.tasks.base_task import TaskBase
from trestle.tasks.base_task import TaskOutcome

HEADER_DECORATION_CHAR = '$'
COMPONENT_DESCRIPTION = 'Component_Description'
COMPONENT_TITLE = 'Component_Title'
COMPONENT_TYPE = 'Component_Type'
CONTROL_ID_LIST = 'Control_Id_List'
NAMESPACE = 'Namespace'
RULE_ID = 'Rule_Id'
RULE_DESCRIPTION = 'Rule_Description'
PROFILE_SOURCE = 'Profile_Source'
PROFILE_DESCRIPTION = 'Profile_Description'
CHECK_ID = 'Check_Id'
CHECK_DESCRIPTION = 'Check_Description'
FETCHER_ID = 'Fetcher_Id'
FETCHER_DESCRIPTION = 'Fetcher_Description'
PARAMETER = 'Parameter'
PARAMETER_ID = f'{PARAMETER}_Id'
PARAMETER_DESCRIPTION = f'{PARAMETER}_Description'
PARAMETER_VALUE_DEFAULT = f'{PARAMETER}_Value_Default'
PARAMETER_VALUE_ALTERNATIVES = f'{PARAMETER}_Value_Alternatives'

validation = 'validation'
prefix_rule_set = 'rule_set_'

logger = logging.getLogger(__name__)


def derive_control_id(control_mapping: str) -> str:
    """Derive control id."""
    rval = control_mapping.split('_smt')[0]
    return rval


def derive_part_id(control_mapping: str) -> str:
    """Derive part id."""
    if '_smt.' in control_mapping:
        rval = control_mapping
    else:
        rval = None
    return rval


def etype(target: str) -> str:
    """Get etype."""
    if target:
        return 'invalid'
    else:
        return 'missing'


def row_property_builder(row: int, name: str, value, ns: str, class_: str, remarks: str) -> Property:
    """Row property builder."""
    # name
    try:
        Property(
            name=name,
            value='value',
        )
    except Exception:
        text = f'property for row: {row} name: {name} is {etype(name)}'
        raise RuntimeError(text)
    # value
    try:
        Property(
            name=name,
            value=value,
        )
    except Exception:
        text = f'property for row: {row} value: {value} is {etype(value)}'
        raise RuntimeError(text)
    # ns
    try:
        Property(
            name=name,
            value=value,
            ns=ns,
        )
    except Exception:
        text = f'property for row: {row} ns: {ns} is {etype(ns)}'
        raise RuntimeError(text)
    # class
    try:
        Property(
            name=name,
            value=value,
            class_=class_,
        )
    except Exception:
        text = f'property for row: {row} class: {class_} is {etype(class_)}'
        raise RuntimeError(text)
    # prop
    prop = Property(
        name=name,
        value=value,
        ns=ns,
        class_=class_,
        remarks=remarks,
    )
    return prop


class CsvToOscalComponentDefinition(TaskBase):
    """
    Task to create OSCAL ComponentDefinition json.

    Attributes:
        name: Name of the task.
    """

    name = 'csv-to-oscal-cd'

    def __init__(self, config_object: Optional[configparser.SectionProxy]) -> None:
        """
        Initialize trestle task csv-to-oscal-cd.

        Args:
            config_object: Config section associated with the task.
        """
        super().__init__(config_object)

    def print_info(self) -> None:
        """Print the help string."""
        name = self.name
        oscal_name = 'component_definition'
        #
        logger.info(f'Help information for {name} task.')
        logger.info('')
        logger.info(f'Purpose: From csv produce OSCAL {oscal_name} file.')
        logger.info('')
        logger.info('')
        logger.info(f'Configuration flags sit under [task.{name}]:')
        text1 = '  title                = '
        text2 = '(required) the component definition title.'
        logger.info(text1 + text2)
        text1 = '  version              = '
        text2 = '(required) the component definition version.'
        logger.info(text1 + text2)
        text1 = '  csv-file             = '
        text2 = '(required) the path of the csv file.'
        text3 = ' [1st row are column headings; 2nd row are column descriptions; 3rd row and beyond is data]'
        logger.info(text1 + text2 + text3)
        text1 = '  required columns:      '
        for text2 in CsvColumn.get_required_column_names():
            if text2 in [f'{RULE_DESCRIPTION}', f'{PROFILE_SOURCE}', f'{PROFILE_DESCRIPTION}', f'{CONTROL_ID_LIST}']:
                text2 += ' (see note 1)'
            logger.info(text1 + '$$' + text2)
            text1 = '                         '
        text1 = '  optional columns:      '
        for text2 in CsvColumn.get_optional_column_names():
            text2 += ' (see note 2)'
            logger.info(text1 + '$' + text2)
            text1 = '                         '
        for text2 in CsvColumn.get_parameter_column_names():
            text2 += ' (see notes 1, 4)'
            logger.info(text1 + '$' + text2)
            text1 = '                         '
        text1 = '  comment columns:       '
        text2 = 'Informational (see note 3)'
        logger.info(text1 + '#' + text2)
        text1 = '  output-dir           = '
        text2 = '(required) the path of the output directory for synthesized OSCAL .json files.'
        logger.info(text1 + text2)
        text1 = '  component-definition = '
        text2 = '(optional) the path of the existing component-definition OSCAL .json file.'
        logger.info(text1 + text2)
        text1 = '  class.column-name    = '
        text2 = f'(optional) the class to associate with the specified column name, e.g. class.{RULE_ID} = scc_class'
        logger.info(text1 + text2)
        text1 = '  output-overwrite     = '
        text2 = '(optional) true [default] or false; replace existing output when true.'
        logger.info(text1 + text2)
        text1 = '  validate-controls    = '
        text2 = '(optional) on, warn, or off [default]; validate controls exist in resolved profile.'
        logger.info(text1 + text2)
        # Notes
        text1 = ''
        text2 = ''
        logger.info(text1 + text2)
        text1 = 'Notes: '
        text2 = '[1] column is ignored for validation component type'
        logger.info(text1 + text2)
        text1 = '       '
        text2 = '[2] column is required for validation component type'
        logger.info(text1 + text2)
        text1 = '       '
        text2 = '[3] column name starting with # causes column to be ignored'
        logger.info(text1 + text2)
        text1 = '       '
        text2 = '[4] additional parameters are specified by adding a common suffix per set'
        text3 = f', for example: {PARAMETER_ID}_1, {PARAMETER_DESCRIPTION}_1, ...{PARAMETER_ID}_2...'
        logger.info(text1 + text2 + text3)

    def configure(self) -> bool:
        """Configure."""
        self._timestamp = datetime.datetime.utcnow().replace(microsecond=0).replace(tzinfo=datetime.timezone.utc
                                                                                    ).isoformat()
        # config verbosity
        self._quiet = self._config.get('quiet', False)
        self._verbose = not self._quiet
        # title
        self._title = self._config.get('title')
        if self._title is None:
            logger.warning('config missing "title"')
            return False
        # version
        self._version = self._config.get('version')
        if self._version is None:
            logger.warning('config missing "version"')
            return False
        # config csv
        self._csv_file = self._config.get('csv-file')
        if self._csv_file is None:
            logger.warning('config missing "csv-file"')
            return False
        self._csv_path = pathlib.Path(self._csv_file)
        if not self._csv_path.exists():
            logger.warning('"csv-file" not found')
            return False
        # announce csv
        if self._verbose:
            logger.info(f'input: {self._csv_file}')
        # config cd
        self._cd_path = None
        self._cd_file = self._config.get('component-definition')
        if self._cd_file is not None:
            self._cd_path = pathlib.Path(self._cd_file)
            if not self._cd_path.exists():
                logger.warning('"component-definition" not found')
                return False
        # workspace
        self._workspace = os.getcwd()
        # validate_controls
        self._validate_controls = self._config.get('validate-controls', 'off')
        return True

    def get_class(self, name: str) -> str:
        """Get class value for specified name from config."""
        key = f'class.{name}'
        return self._config.get(key)

    def simulate(self) -> TaskOutcome:
        """Provide a simulated outcome."""
        return TaskOutcome('simulated-success')

    def execute(self) -> TaskOutcome:
        """Provide an executed outcome."""
        try:
            return self._execute()
        except Exception:
            logger.error(traceback.format_exc())
            return TaskOutcome('failure')

    def _execute(self) -> TaskOutcome:
        """Execute path core."""
        if not self.configure():
            return TaskOutcome('failure')
        # config output
        odir = self._config.get('output-dir')
        opth = pathlib.Path(odir)
        self._overwrite = self._config.getboolean('output-overwrite', True)
        # insure output dir exists
        opth.mkdir(exist_ok=True, parents=True)
        # calculate output file name & check writability
        oname = 'component-definition.json'
        ofile = opth / oname
        if not self._overwrite and pathlib.Path(ofile).exists():
            logger.warning(f'output: {ofile} already exists')
            return TaskOutcome('failure')
        # fetch existing component-definition, if any
        self._cd_mgr = _CdMgr(self._cd_path, self._title, self._timestamp, self._version)
        # fetch csv
        self._csv_mgr = _CsvMgr(self._csv_path)
        # create resolved profile -> catalog helper
        profile_list = self._csv_mgr.get_profile_list()
        self._resolved_profile_catalog_helper = _ResolvedProfileCatalogHelper(profile_list, self._workspace)
        self._unresolved_controls = []
        # calculate deletion, addition & modification rule lists
        rules = self._calculate_rules()
        # calculate deletion, addition & modification set-parameter lists
        set_params = self._calculate_set_params(rules[2])
        # calculate deletion, addition & modification control mapping lists
        control_mappings = self._calculate_control_mappings(rules[2])
        # rule set manager
        self._rule_set_id_mgr = _RuleSetIdMgr(self._cd_mgr.get_max_rule_set_number(), len(rules[1]))
        # rule additions, deletions & modifications (by row)
        self.rules_del(rules[0])
        self.rules_add(rules[1])
        self.rules_mod(rules[2])
        # set-parameters additions, deletions & modifications (by row)
        self.set_params_del(set_params[0])
        self.set_params_add(set_params[1])
        self.set_params_mod(set_params[2])
        # control mappings additions, deletions & modifications (by row)
        self.control_mappings_del(control_mappings[0])
        self.control_mappings_add(control_mappings[1])
        # note: control mappings mod is currently not possible
        # note: add/del user columns not currently supported
        if len(self._unresolved_controls) > 0:
            text = f'Unresolved controls: {self._unresolved_controls}'
            if self._validate_controls == 'warn':
                logger.warn(text)
            elif self._validate_controls == 'on':
                raise RuntimeError(text)
        # prepare new/revised component definition
        component_definition = self._cd_mgr.get_component_definition()
        # write OSCAL ComponentDefinition to file
        if self._verbose:
            logger.info(f'output: {ofile}')
        component_definition.oscal_write(pathlib.Path(ofile))
        return TaskOutcome('success')

    def _calculate_rules(self) -> tuple:
        """Calculate rules add, delete, modify."""
        cd_rules = self._cd_mgr.get_rule_keys()
        csv_rules = self._csv_mgr.get_rule_keys()
        del_rules = []
        add_rules = []
        mod_rules = []
        for key in cd_rules:
            if key in csv_rules:
                continue
            else:
                del_rules.append(key)
                logger.debug(f'rules del: {key}')
        for key in csv_rules:
            if key in cd_rules:
                mod_rules.append(key)
                logger.debug(f'rules mod: {key}')
            else:
                add_rules.append(key)
                logger.debug(f'rules add: {key}')
        return (del_rules, add_rules, mod_rules)

    def _calculate_set_params(self, mod_rules: List) -> tuple:
        """Calculate set parameters add, delete, modify."""
        cd_set_params = self._cd_mgr.get_set_params_keys()
        csv_set_params = self._csv_mgr.get_set_params_keys()
        del_set_params = []
        add_set_params = []
        mod_set_params = []
        for key in cd_set_params:
            rule_key = (key[0], key[1], key[2])
            if rule_key not in mod_rules:
                continue
            if key in csv_set_params:
                continue
            else:
                del_set_params.append(key)
                logger.debug(f'params del: {key}')
        for key in csv_set_params:
            rule_key = (key[0], key[1], key[2])
            if rule_key not in mod_rules:
                continue
            if key in cd_set_params:
                mod_set_params.append(key)
                logger.debug(f'params mod: {key}')
            else:
                add_set_params.append(key)
                logger.debug(f'params add: {key}')
        return (del_set_params, add_set_params, mod_set_params)

    def _calculate_control_mappings(self, mod_rules: List) -> tuple:
        """Calculate control mappings add, delete, modify."""
        cd_controls = self._cd_mgr.get_control_keys()
        csv_controls = self._csv_mgr.get_control_keys()
        del_control_mappings = []
        add_control_mappings = []
        mod_control_mappings = []
        for key in cd_controls:
            rule_key = (key[0], key[1], key[2])
            if rule_key not in mod_rules:
                continue
            if key in csv_controls:
                continue
            else:
                del_control_mappings.append(key)
                logger.debug(f'ctl-maps del: {key}')
        for key in csv_controls:
            rule_key = (key[0], key[1], key[2])
            if rule_key not in mod_rules:
                continue
            if key in cd_controls:
                mod_control_mappings.append(key)
                logger.debug(f'ctl-maps mod: {key}')
            else:
                add_control_mappings.append(key)
                logger.debug(f'ctl-maps add: {key}')
        return (del_control_mappings, add_control_mappings, mod_control_mappings)

    def _get_namespace(self, rule_key: tuple) -> str:
        """Get namespace."""
        return self._csv_mgr.get_value(rule_key, NAMESPACE).strip()

    def _get_prop_name(self, column_name: str) -> str:
        """Get property name."""
        return column_name.lstrip('$')

    def rules_del(self, del_rules: List[str]) -> None:
        """Delete rules."""
        for tokens in del_rules:
            component_title = tokens[0]
            component_type = tokens[1]
            rule_id = tokens[2]
            description = ''
            # component
            component = self._cd_mgr.get_component(component_title, component_type, description)
            # props
            component.props = self._delete_rule_props(component, rule_id)

    def _delete_rule_props(self, component: DefinedComponent, rule_id: str) -> List[Property]:
        """Delete rule props."""
        props = []
        rule_set = _RuleSetHelper.get_rule_set(component.props, rule_id)
        for prop in component.props:
            if prop.remarks != rule_set:
                props.append(prop)
            elif prop.name in self._csv_mgr.get_parameter_id_column_names():
                self._delete_rule_set_parameter(component, prop.value)
            elif prop.name == RULE_ID:
                self._delete_rule_implemented_requirement(component, prop.value)
        return props

    def _control_implementation_generator(
        self, control_implementations: List[ControlImplementation]
    ) -> Iterator[ControlImplementation]:
        """Control implementation generator."""
        if control_implementations:
            for control_implementation in control_implementations:
                yield control_implementation

    def _set_parameter_generator(self, set_parameters: List[SetParameter]) -> Iterator[SetParameter]:
        """Set parameter generator."""
        if set_parameters:
            for set_parameter in set_parameters:
                yield set_parameter

    def _implemented_requirement_generator(
        self, implemented_requirements: List[ImplementedRequirement]
    ) -> Iterator[ImplementedRequirement]:
        """Implemented-requirement generator."""
        if implemented_requirements:
            for implemented_requirement in implemented_requirements:
                yield implemented_requirement

    def _delete_rule_set_parameter(self, component: DefinedComponent, parameter_id: str) -> None:
        """Delete rule set-parameter."""
        control_implementations = component.control_implementations
        for control_implementation in self._control_implementation_generator(control_implementations):
            if control_implementation.set_parameters:
                set_parameters = control_implementation.set_parameters
                control_implementation.set_parameters = []
                for set_parameter in set_parameters:
                    if set_parameter.param_id != parameter_id:
                        _OscalHelper.add_set_parameter(control_implementation.set_parameters, set_parameter)
                if not len(control_implementation.set_parameters):
                    control_implementation.set_parameters = None

    def _delete_rule_implemented_requirement(self, component: DefinedComponent, rule_id: str) -> None:
        """Delete rule implemented_requirement."""
        control_implementations = component.control_implementations
        component.control_implementations = []
        for control_implementation in self._control_implementation_generator(control_implementations):
            if control_implementation.implemented_requirements:
                implemented_requirements = control_implementation.implemented_requirements
                control_implementation.implemented_requirements = []
                for implemented_requirement in implemented_requirements:
                    self._delete_ir_props(implemented_requirement, rule_id)
                    self._delete_ir_statements(implemented_requirement, rule_id)
                    if len(as_list(implemented_requirement.props)) or len(as_list(implemented_requirement.statements)):
                        control_implementation.implemented_requirements.append(implemented_requirement)
            if len(as_list(control_implementation.implemented_requirements)):
                component.control_implementations.append(control_implementation)

    def _delete_ir_statements(self, implemented_requirement: ImplementedRequirement, rule_id: str) -> None:
        """Delete implemented-requirement statements."""
        if implemented_requirement.statements:
            statements = implemented_requirement.statements
            implemented_requirement.statements = []
            for statement in statements:
                statement.props = self._delete_props(statement.props, rule_id)
                if not len(statement.props):
                    statement.props = None
                if statement.props:
                    implemented_requirement.statements.append(statement)
            if not len(implemented_requirement.statements):
                implemented_requirement.statements = None

    def _delete_ir_props(self, implemented_requirement: ImplementedRequirement, rule_id: str) -> None:
        """Delete implemented-requirement props."""
        if implemented_requirement.props:
            implemented_requirement.props = self._delete_props(implemented_requirement.props, rule_id)
            if not len(implemented_requirement.props):
                implemented_requirement.props = None

    def _delete_props(self, props: List[Property], rule_id: str) -> List[property]:
        """Delete props."""
        rval = []
        if props:
            for prop in props:
                if prop.name == RULE_ID and prop.value == rule_id:
                    continue
                rval.append(prop)
        return rval

    def rules_add(self, add_rules: List[str]) -> None:
        """Add rules."""
        for rule_key in add_rules:
            component_title = self._csv_mgr.get_value(rule_key, COMPONENT_TITLE)
            component_type = self._csv_mgr.get_value(rule_key, COMPONENT_TYPE)
            component_description = self._csv_mgr.get_value(rule_key, COMPONENT_DESCRIPTION)
            # component
            component = self._cd_mgr.get_component(component_title, component_type, component_description)
            # props
            component.props = as_list(component.props)
            component.props = component.props + self._create_rule_props(rule_key)
            # additional props, when not validation component
            if not self._is_validation(rule_key):
                # control implementation
                source = self._csv_mgr.get_value(rule_key, PROFILE_SOURCE)
                description = self._csv_mgr.get_value(rule_key, PROFILE_DESCRIPTION)
                control_implementation = self._get_control_implementation(component, source, description)
                # set-parameters
                set_parameters = self._create_set_parameters(rule_key)
                if set_parameters:
                    control_implementation.set_parameters = as_list(control_implementation.set_parameters)
                    _OscalHelper.add_set_parameters(control_implementation.set_parameters, set_parameters)
                # control-mappings
                control_mappings = self._csv_mgr.get_value(rule_key, CONTROL_ID_LIST).split()
                self._add_rule_prop(control_implementation, control_mappings, rule_key)

    def _is_validation(self, rule_key: tuple) -> bool:
        """Check for validation component."""
        component_type = self._csv_mgr.get_value(rule_key, COMPONENT_TYPE)
        return component_type.lower() == validation

    def _add_rule_prop(
        self, control_implementation: ControlImplementation, control_mappings: List[str], rule_key: tuple
    ) -> None:
        """Add rule prop."""
        namespace = self._get_namespace(rule_key)
        for control_mapping in control_mappings:
            control_id = derive_control_id(control_mapping)
            implemented_requirement = self._get_implemented_requirement(control_implementation, control_id)
            # create rule implementation (as property)
            name = RULE_ID
            prop = Property(
                name=name,
                value=self._csv_mgr.get_value(rule_key, name),
                ns=namespace,
                class_=self.get_class(name),
            )
            part_id = derive_part_id(control_mapping)
            if part_id is None:
                implemented_requirement.props = as_list(implemented_requirement.props)
                implemented_requirement.props.append(prop)
            else:
                statement = self._get_statement(implemented_requirement, part_id)
                statement.props.append(prop)

    def _create_rule_props(self, rule_key: tuple) -> List[Property]:
        """Create rule props."""
        rule_set = self._rule_set_id_mgr.get_next_rule_set_id()
        row_number = self._csv_mgr.get_row_number(rule_key)
        rule_set_mgr = _RuleSetMgr(row_number, rule_set)
        namespace = self._get_namespace(rule_key)
        if self._is_validation(rule_key):
            column_names = CsvColumn.get_check_property_column_names()
        else:
            column_names = CsvColumn.get_rule_property_column_names()
        # req'd & optional props
        for column_name in column_names:
            prop_name = self._get_prop_name(column_name)
            prop_value = self._csv_mgr.get_value(rule_key, column_name).strip()
            rule_set_mgr.add_prop(prop_name, prop_value, namespace, self.get_class(prop_name))
        if not self._is_validation(rule_key):
            # parameter columns
            column_names = self._csv_mgr.get_parameter_column_names()
            for column_name in column_names:
                prop_name = self._get_prop_name(column_name)
                prop_value = self._csv_mgr.get_value(rule_key, column_name).strip()
                rule_set_mgr.add_prop(prop_name, prop_value, namespace, self.get_class(prop_name))
        # user props
        column_names = self._csv_mgr.get_user_column_names()
        for column_name in column_names:
            if column_name.startswith('#'):
                continue
            prop_name = self._get_prop_name(column_name)
            prop_value = self._csv_mgr.get_value(rule_key, column_name).strip()
            rule_set_mgr.add_prop(prop_name, prop_value, namespace, self.get_class(prop_name))
        return rule_set_mgr.get_props()

    def _get_control_implementation(
        self, component: DefinedComponent, source: str, description: str
    ) -> ControlImplementation:
        """Find or create control implementation."""
        component.control_implementations = as_list(component.control_implementations)
        for control_implementation in component.control_implementations:
            if control_implementation.source == source and control_implementation.description == description:
                return control_implementation
        control_implementation = ControlImplementation(
            uuid=str(uuid.uuid4()), source=source, description=description, implemented_requirements=[]
        )
        component.control_implementations.append(control_implementation)
        return control_implementation

    def _str_to_list(self, value: str) -> List[str]:
        """Transform string to list."""
        rval = []
        if ',' in value:
            values = value.split(',')
            # remove leading/trailing whitespace
            for v in values:
                rval.append(v.strip())
        else:
            rval.append(value)
        return rval

    def _create_set_parameters(self, rule_key: tuple) -> List[SetParameter]:
        """Create set parameters."""
        set_parameters = []
        for parameter_id_column_name in self._csv_mgr.get_parameter_id_column_names():
            suffix = parameter_id_column_name.replace(PARAMETER_ID, '')
            parameter_value_default_column_name = f'{PARAMETER_VALUE_DEFAULT}{suffix}'
            name = self._csv_mgr.get_value(rule_key, parameter_id_column_name)
            value = self._csv_mgr.get_value(rule_key, parameter_value_default_column_name)
            if name and value:
                values = self._str_to_list(value)
                set_parameter = SetParameter(
                    param_id=name,
                    values=values,
                )
                set_parameters.append(set_parameter)
            elif name:
                row_number = self._csv_mgr.get_row_number(rule_key)
                text = f'row "{row_number}" missing value for "{parameter_value_default_column_name}"'
                logger.debug(text)
        return set_parameters

    def _get_implemented_requirement(
        self, control_implementation: ControlImplementation, control_id: str
    ) -> ImplementedRequirement:
        """Find or create implemented requirement."""
        if self._validate_controls != 'off':
            if not self._resolved_profile_catalog_helper.validate(control_id):
                if control_id not in self._unresolved_controls:
                    self._unresolved_controls.append(control_id)
        for implemented_requirement in control_implementation.implemented_requirements:
            if implemented_requirement.control_id == control_id:
                return implemented_requirement
        implemented_requirement = ImplementedRequirement(
            uuid=str(uuid.uuid4()),
            control_id=control_id,
            description='',
        )
        control_implementation.implemented_requirements.append(implemented_requirement)
        return implemented_requirement

    def _get_statement(self, implemented_requirement: ImplementedRequirement, part_id: str) -> Statement:
        """Find or create statement."""
        implemented_requirement.statements = as_list(implemented_requirement.statements)
        for statement in implemented_requirement.statements:
            if statement.statement_id == part_id:
                return statement
        statement = Statement(
            uuid=str(uuid.uuid4()),
            statement_id=part_id,
            description='',
            props=[],
        )
        implemented_requirement.statements.append(statement)
        return statement

    def rules_mod(self, mod_rules: List[str]) -> None:
        """Modify rules."""
        for rule_key in mod_rules:
            component_title = self._csv_mgr.get_value(rule_key, COMPONENT_TITLE)
            component_type = self._csv_mgr.get_value(rule_key, COMPONENT_TYPE)
            component_description = self._csv_mgr.get_value(rule_key, COMPONENT_DESCRIPTION)
            # component
            component = self._cd_mgr.get_component(component_title, component_type, component_description)
            # props
            component.props = self._modify_rule_props(component, rule_key)

    def _modify_rule_props(self, component: DefinedComponent, rule_key: tuple) -> List[Property]:
        """Modify rule props."""
        rule_id = self._csv_mgr.get_value(rule_key, RULE_ID)
        rule_set = _RuleSetHelper.get_rule_set(component.props, rule_id)
        rule_ns = self._csv_mgr.get_value(rule_key, NAMESPACE)
        column_names = CsvColumn.get_filtered_required_column_names() + CsvColumn.get_filtered_optional_column_names()
        # req'd & optional props
        for column_name in column_names:
            column_value = self._csv_mgr.get_value(rule_key, column_name).strip()
            class_ = self.get_class(column_name)
            self._cd_mgr.update_rule_definition(component, rule_set, column_name, column_value, rule_ns, class_)
        # parameter columns
        column_names = self._csv_mgr.get_parameter_column_names()
        for column_name in column_names:
            column_value = self._csv_mgr.get_value(rule_key, column_name).strip()
            class_ = self.get_class(column_name)
            self._cd_mgr.update_rule_definition(component, rule_set, column_name, column_value, rule_ns, class_)
        # user props
        column_names = self._csv_mgr.get_user_column_names()
        for column_name in column_names:
            column_value = self._csv_mgr.get_value(rule_key, column_name).strip()
            self._cd_mgr.update_rule_definition(component, rule_set, column_name, column_value, rule_ns, class_)
        return component.props

    def set_params_del(self, del_set_params: List[str]) -> None:
        """Set parameters delete."""
        for tokens in del_set_params:
            component_title = tokens[0]
            component_type = tokens[1]
            source = tokens[3]
            description = tokens[4]
            param_id = tokens[5]
            control_implementation = self._cd_mgr.find_control_implementation(
                component_title, component_type, source, description
            )
            if control_implementation:
                set_parameters = control_implementation.set_parameters
                control_implementation.set_parameters = []
                for set_parameter in self._set_parameter_generator(set_parameters):
                    if set_parameter.param_id == param_id:
                        continue
                    _OscalHelper.add_set_parameter(control_implementation.set_parameters, set_parameter)
                if control_implementation.set_parameters == []:
                    control_implementation.set_parameters = None

    def set_params_add(self, add_set_params: List[str]) -> None:
        """Set parameters add."""
        for tokens in add_set_params:
            component_title = tokens[0]
            component_type = tokens[1]
            rule_id = tokens[2]
            source = tokens[3]
            description = tokens[4]
            param_id = tokens[5]
            control_implementation = self._cd_mgr.find_control_implementation(
                component_title, component_type, source, description
            )
            control_implementation.set_parameters = as_list(control_implementation.set_parameters)
            # add
            rule_key = _CsvMgr.get_rule_key(component_title, component_type, rule_id)
            values = [self._csv_mgr.get_default_value_by_id(rule_key, param_id)]
            set_parameter = SetParameter(
                param_id=param_id,
                values=values,
            )
            _OscalHelper.add_set_parameter(control_implementation.set_parameters, set_parameter)

    def set_params_mod(self, mod_set_params: List[str]) -> None:
        """Set parameters modify."""
        for tokens in mod_set_params:
            component_title = tokens[0]
            component_type = tokens[1]
            rule_id = tokens[2]
            source = tokens[3]
            description = tokens[4]
            param_id = tokens[5]
            control_implementation = self._cd_mgr.find_control_implementation(
                component_title, component_type, source, description
            )
            if control_implementation:
                set_parameters = control_implementation.set_parameters
                for set_parameter in self._set_parameter_generator(set_parameters):
                    if set_parameter.param_id != param_id:
                        continue
                    rule_key = _CsvMgr.get_rule_key(component_title, component_type, rule_id)
                    values = [self._csv_mgr.get_default_value_by_id(rule_key, param_id)]
                    replacement = SetParameter(
                        param_id=param_id,
                        values=values,
                    )
                    if set_parameter.values == replacement.values:
                        continue
                    logger.debug(f'params-mod: {rule_id} {param_id} {set_parameter.values} -> {replacement.values}')
                    set_parameter.values = replacement.values

    def _control_mappings_generator(self, control_mappings: List[str]) -> Iterator[List[str]]:
        """Control mappings generator."""
        for tokens in control_mappings:
            component_title = tokens[0]
            component_type = tokens[1]
            source = tokens[3]
            description = tokens[4]
            control_implementation = self._cd_mgr.find_control_implementation(
                component_title, component_type, source, description
            )
            if control_implementation:
                yield tokens

    def control_mappings_del(self, del_control_mappings: List[str]) -> None:
        """Control mappings delete."""
        for tokens in self._control_mappings_generator(del_control_mappings):
            component_title = tokens[0]
            component_type = tokens[1]
            rule_id = tokens[2]
            source = tokens[3]
            description = tokens[4]
            smt_id = tokens[5]
            control_id = derive_control_id(smt_id)
            control_implementation = self._cd_mgr.find_control_implementation(
                component_title, component_type, source, description
            )
            implemented_requirements = control_implementation.implemented_requirements
            control_implementation.implemented_requirements = []
            for implemented_requirement in self._implemented_requirement_generator(implemented_requirements):
                if implemented_requirement.control_id == control_id:
                    implemented_requirement.statements = _OscalHelper.remove_rule_statement(
                        implemented_requirement.statements, rule_id, smt_id
                    )
                    implemented_requirement.props = _OscalHelper.remove_rule(implemented_requirement.props, rule_id)
                    if len(as_list(implemented_requirement.props)) or len(as_list(implemented_requirement.statements)):
                        control_implementation.implemented_requirements.append(implemented_requirement)
                else:
                    control_implementation.implemented_requirements.append(implemented_requirement)

    def control_mappings_add(self, add_control_mappings: List[str]) -> None:
        """Control mappings add."""
        for tokens in self._control_mappings_generator(add_control_mappings):
            component_title = tokens[0]
            component_type = tokens[1]
            rule_id = tokens[2]
            source = tokens[3]
            description = tokens[4]
            smt_id = tokens[5]
            control_id = derive_control_id(smt_id)
            control_implementation = self._cd_mgr.find_control_implementation(
                component_title, component_type, source, description
            )
            implemented_requirement = self._get_implemented_requirement(control_implementation, control_id)
            # namespace
            rule_key = (tokens[0], tokens[1], tokens[2])
            ns = self._get_namespace(rule_key)
            # create rule implementation (as property)
            name = RULE_ID
            prop = Property(
                name=name,
                value=rule_id,
                ns=ns,
                class_=self.get_class(name),
            )
            if smt_id == control_id:
                implemented_requirement.props = as_list(implemented_requirement.props)
                implemented_requirement.props.append(prop)
            else:
                statement = self._get_statement(implemented_requirement, smt_id)
                statement.props.append(prop)


class _OscalHelper():
    """Oscal Helper."""

    @staticmethod
    def add_set_parameters(set_parameter_list: List[SetParameter], set_parameter_list_add: List[SetParameter]) -> None:
        """Add set parameters."""
        for set_parameter in set_parameter_list_add:
            _OscalHelper.add_set_parameter(set_parameter_list, set_parameter)

    @staticmethod
    def add_set_parameter(set_parameter_list: List[SetParameter], set_parameter: SetParameter) -> None:
        """Add set parameter."""
        add = True
        # don't add duplicate
        for sp in set_parameter_list:
            if sp.param_id == set_parameter.param_id:
                add = False
                if sp.values != set_parameter.values:
                    text = f'set-parameter id={sp.param_id} conflicting values'
                    raise RuntimeError(text)
                break
        if add:
            set_parameter_list.append(set_parameter)

    @staticmethod
    def remove_rule_statement(statements: List[Statement], rule_id: str, smt_id: str) -> List[Statement]:
        """Remove rule from statements."""
        rval = statements
        if statements:
            rval = []
            for statement in statements:
                if statement.statement_id == smt_id:
                    statement.props = _OscalHelper.remove_rule(statement.props, rule_id)
                if statement.props is not None and len(statement.props):
                    rval.append(statement)
        return rval

    @staticmethod
    def remove_rule(props: List[Property], rule_id: str) -> List[Property]:
        """Remove rule from props."""
        rval = props
        if props:
            rval = []
            for prop in props:
                if prop.name != RULE_ID or prop.value != rule_id:
                    rval.append(prop)
        return rval


class _RuleSetHelper():
    """RuleSet Helper."""

    @staticmethod
    def get_rule_set(props: List[Property], rule_id: str) -> str:
        """Get rule_set for given rule_id."""
        rule_set = None
        if props:
            for prop in props:
                if prop.name == 'Rule_Id' and prop.value == rule_id:
                    rule_set = prop.remarks
                    break
        return rule_set


class _RuleSetIdMgr():
    """RuleSetId Manager."""

    def __init__(self, max_rule_set_number: int, add_rules_count: int) -> None:
        """Initialize."""
        self._prev_rule_set_number = max_rule_set_number
        self._rule_set_number_digits = max_rule_set_number + add_rules_count + 1
        self._fill_sz = int(log10(self._rule_set_number_digits)) + 1

    def get_next_rule_set_id(self) -> str:
        self._prev_rule_set_number += 1
        rval = f'{prefix_rule_set}{str(self._prev_rule_set_number).zfill(self._fill_sz)}'
        return rval


class _RuleSetMgr():
    """RuleSet manager."""

    def __init__(self, row_number: int, rule_set: str) -> None:
        """Initialize."""
        self._row_number = row_number
        self._rule_set = rule_set
        self._props = {}

    def add_prop(self, name: str, value: str, ns: str, class_: str) -> None:
        """Add prop."""
        if value is not None and len(value):
            prop = row_property_builder(
                row=self._row_number,
                name=name,
                value=value,
                ns=ns,
                class_=class_,
                remarks=self._rule_set,
            )
            self._props[name] = prop

    def get_props(self) -> List[Property]:
        """Get props."""
        rval = []
        # required
        c1 = CsvColumn.get_required_column_names()
        for key in c1:
            if key in self._props.keys():
                rval.append(self._props[key])
        # parameter
        c3 = []
        for key in self._props.keys():
            if key.startswith(f'{PARAMETER}'):
                rval.append(self._props[key])
                c3.append(key)
        # optional
        c2 = CsvColumn.get_optional_column_names()
        for key in c2:
            if key in self._props.keys():
                rval.append(self._props[key])
        # user
        for key in self._props.keys():
            if key in c1 or key in c2 or key in c3:
                continue
            rval.append(self._props[key])
        return rval


class _ResolvedProfileCatalogHelper():
    """Resolved Profile Catalog Helper."""

    def __init__(self, profile_list: List[str], root: str = '.') -> None:
        """Initialize."""
        self._profile_list = profile_list
        self._root = root
        self._profile_map = {}
        self._control_list = []
        self._init = False

    def _initialize(self):
        if not self._init:
            for profile in self._profile_list:
                catalog = ProfileResolver.get_resolved_profile_catalog(
                    pathlib.Path(self._root),
                    profile,
                )
                self._profile_map[profile] = catalog
                controls = CatalogInterface.get_control_ids_from_catalog(catalog)
                self._control_list += controls
            logger.debug(f'resolved controls: {self._control_list}')
            self._init = True

    def validate(self, control_id: str) -> bool:
        """Validate control_id."""
        self._initialize()
        rval = True
        if control_id not in self._control_list:
            rval = False
        return rval


class _CdMgr():
    """CD Manager."""

    def __init__(self, cd_path: pathlib.Path, title: str, timestamp: str, version: str) -> None:
        """Initialize."""
        if cd_path:
            self._component_definition = ComponentDefinition.oscal_read(cd_path)
            metadata = self._component_definition.metadata
            metadata.title = title
            metadata.last_modified = timestamp
            metadata.oscal_version = OSCAL_VERSION
            metadata.version = version
        else:
            metadata = Metadata(
                title=title,
                last_modified=timestamp,
                oscal_version=OSCAL_VERSION,
                version=version,
            )
            self._component_definition = ComponentDefinition(
                uuid=str(uuid.uuid4()),
                metadata=metadata,
                components=[],
            )
        #
        self._max_rule_set_number = -1
        self._cd_rules_map = {}
        self._cd_set_params_map = {}
        self._cd_controls_map = {}
        #
        for component in self._component_definition.components:
            self.accounting(component)
        logger.debug(f'cd rules: {len(self._cd_rules_map)}')
        logger.debug(f'cd params: {len(self._cd_rules_map)}')
        logger.debug(f'cd controls: {len(self._cd_controls_map)}')

    def get_component(self, component_title: str, component_type: str, component_description: str) -> DefinedComponent:
        """Get component."""
        for component in self._component_definition.components:
            if component.title == component_title and component.type == component_type:
                logger.debug(f'located component: title={component.title} type={component.type}')
                return component
        component = DefinedComponent(
            uuid=str(uuid.uuid4()),
            type=component_type,
            title=component_title,
            description=component_description,
            control_implementations=[],
        )
        self._component_definition.components.append(component)
        logger.debug(f'created component: title={component.title} type={component.type}')
        return component

    def find_component(self, component_title: str, component_type: str) -> DefinedComponent:
        """Find component."""
        rval = None
        for component in self._component_definition.components:
            if component.title == component_title and component.type == component_type:
                logger.debug(f'located component: title={component.title} type={component.type}')
                rval = component
                break
        return rval

    def find_control_implementation(
        self, component_title: str, component_type: str, source: str, description: str
    ) -> ControlImplementation:
        """Find control-implementation."""
        rval = None
        component = self.find_component(component_title, component_type)
        for control_implementation in component.control_implementations:
            if control_implementation.source == source and control_implementation.description == description:
                rval = control_implementation
                break
        return rval

    def get_component_definition(self) -> ComponentDefinition:
        """Get component definition."""
        # remove empty components
        self._remove_empty_components()
        return self._component_definition

    def _remove_empty_components(self) -> None:
        """Remove empty components."""
        component_definition = self._component_definition
        components = component_definition.components
        component_definition.components = []
        for component in components:
            if component.props is None or len(component.props) == 0:
                if component.control_implementations is None or len(component.control_implementations) == 0:
                    continue
            component_definition.components.append(component)

    def get_max_rule_set_number(self) -> int:
        """Get max rule set number."""
        return self._max_rule_set_number

    def get_rule_keys(self) -> List[str]:
        """Get rule keys."""
        return self._cd_rules_map.keys()

    def get_set_params_keys(self) -> List[str]:
        """Get set-parameter keys."""
        return self._cd_set_params_map.keys()

    def get_control_keys(self) -> List[str]:
        """Get control keys."""
        return self._cd_controls_map.keys()

    def accounting(self, component: DefinedComponent) -> None:
        """Accounting."""
        # rule definitions
        self.accounting_rule_definitions(component)
        # set-parameters & control mappings
        if component.control_implementations:
            for control_implementation in component.control_implementations:
                # set-parameters
                self.accounting_set_parameters(component, control_implementation)
                # control mappings
                self.accounting_control_mappings(component, control_implementation)

    def accounting_rule_definitions(self, component: DefinedComponent) -> None:
        """Accounting, rule definitions."""
        if component.props:
            for prop in component.props:
                if prop.name == RULE_ID:
                    key = (component.title, component.type, prop.value)
                    value = prop.remarks
                    self._cd_rules_map[key] = value
                    logger.debug(f'cd: {key} {self._cd_rules_map[key]}')
                    rule_set_number = int(value.replace(f'{prefix_rule_set}', ''))
                    if rule_set_number > self._max_rule_set_number:
                        self._max_rule_set_number = rule_set_number

    def accounting_set_parameters(
        self, component: DefinedComponent, control_implementation: ControlImplementation
    ) -> None:
        """Accounting, set-parameters."""
        if control_implementation.set_parameters:
            for set_parameter in control_implementation.set_parameters:
                rule_id = self._get_rule_id(component, set_parameter.param_id)
                key = (
                    component.title,
                    component.type,
                    rule_id,
                    control_implementation.source,
                    control_implementation.description,
                    set_parameter.param_id
                )
                value = set_parameter.values
                self._cd_set_params_map[key] = value

    def accounting_control_mappings(
        self, component: DefinedComponent, control_implementation: ControlImplementation
    ) -> None:
        """Accounting, control mappings."""
        if control_implementation.implemented_requirements:
            for implemented_requirement in control_implementation.implemented_requirements:
                self.accounting_control_mappings_props(component, control_implementation, implemented_requirement)
                self.accounting_control_mappings_statements(component, control_implementation, implemented_requirement)

    def accounting_control_mappings_props(
        self,
        component: DefinedComponent,
        control_implementation: ControlImplementation,
        implemented_requirement: ImplementedRequirement
    ) -> None:
        """Accounting, control mappings props."""
        if implemented_requirement.props:
            for prop in implemented_requirement.props:
                if prop.name == RULE_ID:
                    rule_id = prop.value
                    key = (
                        component.title,
                        component.type,
                        rule_id,
                        control_implementation.source,
                        control_implementation.description,
                        implemented_requirement.control_id,
                    )
                    self._cd_controls_map[key] = prop

    def accounting_control_mappings_statements(
        self,
        component: DefinedComponent,
        control_implementation: ControlImplementation,
        implemented_requirement: ImplementedRequirement
    ) -> None:
        """Accounting, control mappings statements."""
        if implemented_requirement.statements:
            for statement in implemented_requirement.statements:
                if statement.props:
                    for prop in statement.props:
                        if prop.name == RULE_ID:
                            rule_id = prop.value
                            key = (
                                component.title,
                                component.type,
                                rule_id,
                                control_implementation.source,
                                control_implementation.description,
                                statement.statement_id,
                            )
                            self._cd_controls_map[key] = prop

    def _get_rule_id(self, component: DefinedComponent, param_id: str) -> str:
        """Get rule_id for given param_id."""
        rule_id = None
        if component.props:
            map_ = {}
            rule_set = None
            for prop in component.props:
                if prop.name == 'Rule_Id':
                    map_[prop.remarks] = prop.value
                elif prop.name == 'Parameter_Id' and prop.value == param_id:
                    rule_set = prop.remarks
            if rule_set:
                rule_id = map_[rule_set]
        return rule_id

    def update_rule_definition(
        self, component: DefinedComponent, rule_set: str, name: str, value: str, ns: str, class_: str
    ) -> None:
        """Update rule definition."""
        if value is not None and len(value):
            prop = self.find_property(component, rule_set, name)
            if prop:
                # no change
                if prop.value == value:
                    return
                # replace value
                logger.debug(f'update-rule: {rule_set} {name} {prop.value} -> {value}')
                prop.value = value
            else:
                self.add_property(component, rule_set, name, value, ns, class_)
        else:
            self.delete_property(component, rule_set, name)

    def find_property(self, component: DefinedComponent, rule_set: str, name: str) -> Property:
        """Find property."""
        rval = None
        for prop in component.props:
            if prop.remarks == rule_set and prop.name == name:
                rval = prop
                break
        return rval

    def add_property(
        self, component: DefinedComponent, rule_set: str, name: str, value: str, ns: str, class_: str
    ) -> None:
        """Add property."""
        prop_add = Property(
            name=name,
            value=value,
            ns=ns,
            class_=class_,
            remarks=rule_set,
        )
        last = 0
        for index, prop in enumerate(component.props):
            if prop.remarks == rule_set:
                last = index
        props = []
        for index, prop in enumerate(component.props):
            if prop_add:
                if index > last:
                    props.append(prop_add)
                    prop_add = None
                    logger.debug(f'add-prop (last): {rule_set} {name} {prop.value} ->> {value}')
                elif prop_add.remarks == prop.remarks:
                    if CsvColumn.get_order(prop.name) > CsvColumn.get_order(prop_add.name):
                        props.append(prop_add)
                        prop_add = None
                        logger.debug(f'add-prop (order): {rule_set} {name} {prop.value} ->> {value}')
            props.append(prop)
        component.props = props

    def delete_property(self, component: DefinedComponent, rule_set: str, name: str) -> None:
        """Delete property."""
        props = []
        for prop in component.props:
            if prop.remarks == rule_set and prop.name == name:
                logger.debug(f'delete-prop: {rule_set} {name} {prop.value}')
            else:
                props.append(prop)
        component.props = props


class CsvColumn():
    """CsvColumn."""

    _columns_required = [
        f'{COMPONENT_TITLE}',
        f'{COMPONENT_DESCRIPTION}',
        f'{COMPONENT_TYPE}',
        f'{RULE_ID}',
        f'{RULE_DESCRIPTION}',
        f'{PROFILE_SOURCE}',
        f'{PROFILE_DESCRIPTION}',
        f'{CONTROL_ID_LIST}',
        f'{NAMESPACE}',
    ]

    _columns_required_validation = [
        f'{COMPONENT_TITLE}',
        f'{COMPONENT_DESCRIPTION}',
        f'{COMPONENT_TYPE}',
        f'{RULE_ID}',
        f'{NAMESPACE}',
        f'{CHECK_ID}',
        f'{CHECK_DESCRIPTION}',
    ]

    _columns_optional = [
        f'{CHECK_ID}',
        f'{CHECK_DESCRIPTION}',
    ]

    _columns_parameter = [
        f'{PARAMETER_ID}',
        f'{PARAMETER_DESCRIPTION}',
        f'{PARAMETER_VALUE_ALTERNATIVES}',
        f'{PARAMETER_VALUE_DEFAULT}',
    ]

    _columns_ordered = _columns_required + _columns_parameter + _columns_optional

    @staticmethod
    def get_order(column_name: str) -> int:
        """Get order for column_name."""
        rval = sys.maxsize
        if column_name in CsvColumn._columns_ordered:
            rval = CsvColumn._columns_ordered.index(column_name)
        return rval

    @staticmethod
    def is_column_name_required(name: str) -> bool:
        """Is column name required."""
        return name in (CsvColumn._columns_required + CsvColumn._columns_required_validation)

    @staticmethod
    def is_column_name_optional(name: str) -> bool:
        """Is column name optional."""
        return name in (CsvColumn._columns_optional)

    @staticmethod
    def is_column_name_parameter(name: str) -> bool:
        """Is column name parameter."""
        for cname in CsvColumn._columns_parameter:
            if name.startswith(cname):
                return True
        return False

    @staticmethod
    def get_required_column_names() -> List[str]:
        """Get required column names."""
        rval = []
        rval += CsvColumn._columns_required
        return rval

    @staticmethod
    def get_optional_column_names() -> List[str]:
        """Get optional column names."""
        rval = []
        rval += CsvColumn._columns_optional
        return rval

    @staticmethod
    def get_parameter_column_names() -> List[str]:
        """Get parameter column names."""
        rval = []
        rval += CsvColumn._columns_parameters
        return rval

    @staticmethod
    def get_required_column_names_validation() -> List[str]:
        """Get required column names validation."""
        rval = []
        rval += CsvColumn._columns_required_validation
        return rval

    _rule_property_column_names = [
        f'{RULE_ID}',
        f'{RULE_DESCRIPTION}',
        f'{PARAMETER_ID}',
        f'{PARAMETER_DESCRIPTION}',
        f'{PARAMETER_VALUE_ALTERNATIVES}',
        f'{CHECK_ID}',
        f'{CHECK_DESCRIPTION}',
    ]

    @staticmethod
    def get_rule_property_column_names() -> List[str]:
        """Get rule property column names."""
        return CsvColumn._rule_property_column_names

    # columns required which do not become properties
    _columns_required_filtered = [
        f'{COMPONENT_TITLE}',
        f'{COMPONENT_DESCRIPTION}',
        f'{COMPONENT_TYPE}',
        f'{PROFILE_SOURCE}',
        f'{PROFILE_DESCRIPTION}',
        f'{CONTROL_ID_LIST}',
        f'{NAMESPACE}',
    ]

    # optional columns which do not become properties, initially
    _columns_optional_filtered = []

    _columns_filtered = _columns_required_filtered + _columns_optional_filtered

    @staticmethod
    def get_filtered_required_column_names() -> List[str]:
        """Get filtered required column names."""
        rval = []
        for column_name in CsvColumn.get_required_column_names():
            if column_name not in CsvColumn._columns_filtered:
                rval.append(column_name)
        return rval

    @staticmethod
    def get_filtered_optional_column_names() -> List[str]:
        """Get filtered optional column names."""
        rval = []
        for column_name in CsvColumn.get_optional_column_names():
            if column_name not in CsvColumn._columns_filtered:
                rval.append(column_name)
        return rval

    _check_property_column_names = [
        f'{RULE_ID}',
        f'{CHECK_ID}',
        f'{CHECK_DESCRIPTION}',
    ]

    @staticmethod
    def get_check_property_column_names() -> List[str]:
        """Get check property column names."""
        return CsvColumn._check_property_column_names

    # optional columns which do become properties, afterwards
    _columns_parameters = [
        f'{PARAMETER_ID}',
        f'{PARAMETER_DESCRIPTION}',
        f'{PARAMETER_VALUE_ALTERNATIVES}',
    ]

    # optional columns which require Param_Id be present in the row
    _columns_parameters_dependent = [
        f'{PARAMETER_DESCRIPTION}',
        f'{PARAMETER_VALUE_ALTERNATIVES}',
        f'{PARAMETER_VALUE_DEFAULT}',
    ]


Row = Iterator[List[str]]


class _CsvMgr():
    """Csv Manager."""

    def __init__(self, csv_path: pathlib.Path) -> None:
        """Initialize."""
        self._csv = []
        with open(csv_path, 'r', newline='') as f:
            csv_reader = csv.reader(f, delimiter=',', quoting=csv.QUOTE_MINIMAL)
            for row in csv_reader:
                self._csv.append(row)
        self._undecorate_header()
        self._verify()
        self._csv_rules_map = {}
        self._csv_set_params_map = {}
        self._csv_controls_map = {}
        self._csv_profile_list = []
        for row_num, row in self.row_generator():
            self._check_row_minimum_requirements(row_num, row)
            component_title = self.get_row_value(row, f'{COMPONENT_TITLE}')
            component_type = self.get_row_value(row, f'{COMPONENT_TYPE}')
            component_description = self.get_row_value(row, f'{COMPONENT_DESCRIPTION}')
            rule_id = self.get_row_value(row, f'{RULE_ID}')
            # rule sets
            key = _CsvMgr.get_rule_key(component_description, component_type, rule_id)
            if key in self._csv_rules_map:
                text = f'row "{row_num}" contains duplicate {RULE_ID} "{rule_id}"'
                raise RuntimeError(text)
            self._csv_rules_map[key] = [row_num, row]
            logger.debug(f'csv-rules: {key} {self._csv_rules_map[key][0]}')
            # set parameters, by component
            source = self.get_row_value(row, PROFILE_SOURCE)
            if source and source not in self._csv_profile_list:
                self._csv_profile_list.append(source)
            description = self.get_row_value(row, PROFILE_DESCRIPTION)
            for param_id_key in self.get_parameter_id_column_names():
                param_id = self.get_row_value(row, param_id_key)
                if param_id:
                    key = (component_title, component_type, rule_id, source, description, param_id)
                    self._csv_set_params_map[key] = [row_num, row]
                    logger.debug(f'csv-set-parameters: {key} {self._csv_set_params_map[key][0]}')
            # control mappings
            self._control_mappings(row_num, row, component_description, component_type, rule_id, source, description)
        logger.debug(f'csv rules: {len(self._csv_rules_map)}')
        logger.debug(f'csv params: {len(self._csv_set_params_map)}')
        logger.debug(f'csv controls: {len(self._csv_controls_map)}')

    def _control_mappings(
        self,
        row_num: int,
        row: Row,
        component_description: str,
        component_type: str,
        rule_id: str,
        source: str,
        description: str
    ) -> None:
        """Control_mappings."""
        control_mappings = self.get_row_value(row, CONTROL_ID_LIST)
        if control_mappings:
            controls = control_mappings.split()
            for control in controls:
                key = (component_description, component_type, rule_id, source, description, control)
                self._csv_controls_map[key] = [row_num, row]

    @staticmethod
    def get_rule_key(component_title: str, component_type: str, rule_id: str) -> tuple:
        """Get rule_key."""
        return (component_title, component_type, rule_id)

    def get_parameter_id_column_names(self) -> List[str]:
        """Get parameter_id column_names."""
        col_names = []
        for col_name in self._csv[0]:
            if col_name.startswith(PARAMETER_ID):
                col_names.append(col_name)
        return col_names

    def get_parameter_column_names(self) -> List[str]:
        """Get parameter column_names."""
        col_names = []
        for col_name in self._csv[0]:
            if col_name.startswith(PARAMETER_VALUE_DEFAULT):
                continue
            if col_name.startswith(PARAMETER):
                col_names.append(col_name)
        return col_names

    def get_profile_list(self):
        """Get profile list."""
        return [] + self._csv_profile_list

    def row_generator(self) -> Generator[Union[int, Row], None, None]:
        """Generate rows."""
        index = 0
        for row in self._csv:
            index += 1
            if index < 3:
                continue
            if self._is_no_control(row):
                continue
            logger.debug(f'row_gen: {index} {row}')
            yield index, row

    def _check_row_minimum_requirements(self, row_num: int, row: List) -> None:
        """Check row minimum requirements."""
        if self._is_component_type_validation(row):
            column_names = CsvColumn.get_required_column_names_validation()
        else:
            column_names = CsvColumn.get_required_column_names()
        for column_name in column_names:
            value = self.get_row_value(row, column_name)
            if value is None or value == '':
                text = f'row "{row_num}" missing value for "{column_name}"'
                raise RuntimeError(text)

    def _is_no_control(self, row: List) -> bool:
        """Check for no control."""
        if self._is_component_type_validation(row):
            rval = False
        else:
            control_id_list = self.get_row_value(row, f'{CONTROL_ID_LIST}')
            if control_id_list.strip() == '':
                rval = True
            else:
                rval = False
        return rval

    def _is_component_type_validation(self, row: List) -> bool:
        """Check for component type validation."""
        component_type = self.get_row_value(row, f'{COMPONENT_TYPE}')
        if component_type.lower().strip() == validation:
            rval = True
        else:
            rval = False
        return rval

    def _undecorate_header(self) -> None:
        """Undecorate header."""
        head_row = self._csv[0]
        self._csv[0] = []
        for column_name in head_row:
            heading = self._get_normalized_column_name(column_name)
            self._csv[0].append(heading)

    def _verify(self) -> None:
        """Verify."""
        required_columns = CsvColumn.get_required_column_names()
        if len(self._csv):
            head_row = self._csv[0]
            for heading in head_row:
                if heading in required_columns:
                    required_columns.remove(heading)
        if len(required_columns):
            text = f'Missing columns: {required_columns}'
            raise RuntimeError(text)

    def get_rule_keys(self) -> List[tuple]:
        """Get rule keys."""
        return self._csv_rules_map.keys()

    def get_set_params_keys(self) -> List[tuple]:
        """Get set-parameter keys."""
        return self._csv_set_params_map.keys()

    def get_control_keys(self) -> List[tuple]:
        """Get control keys."""
        return self._csv_controls_map.keys()

    def _get_normalized_column_name(self, column_name: str) -> str:
        """Get normalized column name."""
        return column_name.lstrip('$')

    def get_col_index(self, column_name: str) -> int:
        """Get index for column name."""
        rval = -1
        index = 0
        head_row = self._csv[0]
        col_name = self._get_normalized_column_name(column_name)
        for heading in head_row:
            head_name = self._get_normalized_column_name(heading)
            if head_name == col_name:
                rval = index
                break
            index += 1
        return rval

    def get_row(self, rule_key: tuple) -> List:
        """Get row for rule."""
        return self._csv_rules_map[rule_key][1]

    def get_row_number(self, rule_key: tuple) -> List:
        """Get row number for rule."""
        return self._csv_rules_map[rule_key][0]

    def get_row_value(self, row: List[str], name: str) -> str:
        """Get value for specified name."""
        rval = ''
        index = self.get_col_index(name)
        if index >= 0:
            rval = row[index]
        return rval

    def get_value(self, rule_key: tuple, name: str) -> str:
        """Get value for specified name."""
        row = self.get_row(rule_key)
        return self.get_row_value(row, name)

    def get_default_value_by_id(self, rule_key: tuple, name: str) -> str:
        """Get default value for specified parameter id."""
        rval = None
        for parameter_id_column_name in self.get_parameter_id_column_names():
            parameter_id_name = self.get_value(rule_key, parameter_id_column_name)
            if name == parameter_id_name:
                suffix = parameter_id_column_name.replace(PARAMETER_ID, '')
                parameter_value_default_column_name = f'{PARAMETER_VALUE_DEFAULT}{suffix}'
                rval = self.get_value(rule_key, parameter_value_default_column_name)
                break
        return rval

    def get_user_column_names(self) -> List[str]:
        """Get user column names."""
        user_column_names = []
        for column_name in self._csv[0]:
            t1 = CsvColumn.is_column_name_required(column_name)
            t2 = CsvColumn.is_column_name_optional(column_name)
            t3 = CsvColumn.is_column_name_parameter(column_name)
            if not (t1 or t2 or t3):
                user_column_names.append(column_name)
        return user_column_names



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create resolved catalog from profile."""

import logging
from typing import Iterator, List, Optional

import trestle.oscal.catalog as cat
import trestle.oscal.profile as prof
from trestle.common.common_types import OBT
from trestle.common.const import RESOLUTION_SOURCE, TRESTLE_INHERITED_PROPS_TRACKER
from trestle.common.err import TrestleNotFoundError
from trestle.common.list_utils import as_list, get_item_from_list, none_if_empty
from trestle.core.catalog.catalog_interface import CatalogInterface
from trestle.core.control_interface import ParameterRep
from trestle.core.pipeline import Pipeline
from trestle.oscal import OSCAL_VERSION, common

logger = logging.getLogger(__name__)


class Modify(Pipeline.Filter):
    """Modify the controls based on the profile."""

    def __init__(
        self,
        profile: prof.Profile,
        change_prose: bool = False,
        block_adds: bool = False,
        block_params: bool = False,
        params_format: str = None,
        param_rep: ParameterRep = ParameterRep.LEAVE_MOUSTACHE,
        show_value_warnings: bool = False,
        value_assigned_prefix: Optional[str] = None,
        value_not_assigned_prefix: Optional[str] = None
    ) -> None:
        """Initialize the filter."""
        self._profile = profile
        self._catalog_interface: Optional[CatalogInterface] = None
        self._block_adds = block_adds
        self._block_params = block_params
        self._change_prose = change_prose
        self._params_format = params_format
        self._param_rep = param_rep
        self.show_value_warnings = show_value_warnings
        self._value_assigned_prefix = value_assigned_prefix
        self._value_not_assigned_prefix = value_not_assigned_prefix
        logger.debug(f'modify initialize filter with profile {profile.metadata.title}')

    @staticmethod
    def _add_adds_to_part(part: common.Part, add: prof.Add, added_parts=False) -> None:
        for attr in ['params', 'props', 'parts', 'links']:
            # don't add parts if already added earlier
            if added_parts and attr == 'parts':
                continue
            add_list = getattr(add, attr, None)
            if add_list:
                Modify._add_attr_to_part(part, add_list[:], attr, add.position)

    @staticmethod
    def _add_to_list(parts_list: List[common.Part], add: prof.Add) -> bool:
        """Add the contents of the add according to its by_id and position.

        Return True on success or False if id needed and not found.

        This is only called when by_id is not None.
        The add will be inserted if the id is found, or return False if not.
        This allows a separate recursive routine to search sub-lists for the id.
        Position before/after will put the item adjacent to the target in the same list as target
        Position starting/ending will put the item within the target itself
        """
        # Test here for matched by_id attribute.
        added_parts = False
        for index in range(len(parts_list)):
            # find the matching part
            if parts_list[index].id == add.by_id:
                if add.position == prof.Position.after:
                    for offset, new_item in enumerate(as_list(add.parts)):
                        parts_list.insert(index + 1 + offset, new_item)
                    added_parts = True
                elif add.position == prof.Position.before:
                    for offset, new_item in enumerate(as_list(add.parts)):
                        parts_list.insert(index + offset, new_item)
                    added_parts = True
                # if starting or ending or None, the adds go directly into this part according to type
                Modify._add_adds_to_part(parts_list[index], add, added_parts)
                return True
        return False

    @staticmethod
    def _add_to_parts(parts: List[common.Part], add: prof.Add) -> bool:
        """
        Add the add to the parts.

        This is only called if add.by_id is not None.
        """
        if Modify._add_to_list(parts, add):
            return True
        for part in parts:
            if part.parts is not None and Modify._add_to_parts(part.parts, add):
                return True
        return False

    @staticmethod
    def _add_attr_to_part(part: common.Part, items: List[OBT], attr: str, position: Optional[prof.Position]) -> None:
        attr_list = as_list(getattr(part, attr, None))
        # position may be None and if so will go at end
        if position in [prof.Position.starting, prof.Position.before]:
            items.extend(attr_list)
            attr_list = items
        else:
            attr_list.extend(items)
        setattr(part, attr, attr_list)

    @staticmethod
    def _add_attr_to_control(
        control: cat.Control, items: List[OBT], attr: str, position: Optional[prof.Position]
    ) -> None:
        attr_list = as_list(getattr(control, attr, None))
        # if position is None it will add to end
        if position in [prof.Position.starting, prof.Position.before]:
            items.extend(attr_list)
            attr_list = items
        else:
            attr_list.extend(items)
        setattr(control, attr, attr_list)

    @staticmethod
    def _add_to_trestle_props(control: cat.Control, add: prof.Add) -> None:
        """Add props to special trestle part that keeps track of inherited props."""
        if add.props:
            trestle_part = get_item_from_list(control.parts, TRESTLE_INHERITED_PROPS_TRACKER, lambda p: p.name)
            if trestle_part is None:
                trestle_part = common.Part(
                    id=TRESTLE_INHERITED_PROPS_TRACKER, name=TRESTLE_INHERITED_PROPS_TRACKER, props=[], parts=[]
                )
                control.parts = as_list(control.parts)
                control.parts.append(trestle_part)
                trestle_part = control.parts[-1]
            if add.by_id is None or add.by_id == control.id:
                trestle_part.props.extend(add.props)
            else:
                by_id_part = get_item_from_list(trestle_part.parts, add.by_id, lambda p: p.title)
                if by_id_part is None:
                    trestle_part.parts.append(
                        common.Part(name=TRESTLE_INHERITED_PROPS_TRACKER + '_' + add.by_id, title=add.by_id, props=[])
                    )
                    by_id_part = trestle_part.parts[-1]
                by_id_part.props.extend(add.props)

    @staticmethod
    def _add_to_control(control: cat.Control, add: prof.Add) -> None:
        """First step in applying Add to control."""
        control.parts = as_list(control.parts)
        if add.by_id is None or add.by_id == control.id:
            # add contents will be added to the control directly and with no recursion
            for attr in ['params', 'props', 'parts', 'links']:
                add_list = getattr(add, attr, None)
                if add_list:
                    Modify._add_attr_to_control(control, add_list[:], attr, add.position)
        else:
            # this is only called if by_id is not None
            if not Modify._add_to_parts(control.parts, add):
                logger.warning(f'Could not find id for add in control {control.id}: {add.by_id}')
        Modify._add_to_trestle_props(control, add)
        control.parts = none_if_empty(control.parts)

    @staticmethod
    def _set_overwrite_items(param: common.Parameter, set_param: prof.SetParameter) -> None:
        # these overwrite
        if set_param.class_:
            param.class_ = set_param.class_
        if set_param.depends_on:
            param.depends_on = set_param.depends_on
        if set_param.label:
            param.label = set_param.label
        if set_param.usage:
            param.usage = set_param.usage
        if set_param.values:
            param.values = set_param.values
        if set_param.select:
            param.select = set_param.select

    @staticmethod
    def _set_appended_items(param: common.Parameter, set_param: prof.SetParameter) -> None:
        # these append
        if set_param.constraints:
            if not param.constraints:
                param.constraints = []
            param.constraints.extend(set_param.constraints)
        if set_param.guidelines:
            if not param.guidelines:
                param.guidelines = []
            param.guidelines.extend(set_param.guidelines)

    @staticmethod
    def _set_replaced_or_appended_items(param: common.Parameter, set_param: prof.SetParameter) -> None:
        # these replace or append
        if set_param.props:
            new_props = as_list(param.props)
            names = [prop.name for prop in new_props]
            for prop in set_param.props:
                if prop.name in names:
                    new_props[names.index(prop.name)] = prop
                else:
                    new_props.append(prop)
            param.props = new_props
        if set_param.links:
            new_links = as_list(param.links)
            hrefs = [link.href for link in new_links]
            for link in set_param.links:
                if link.href in hrefs:
                    new_links[hrefs.index(link.href)] = link
                else:
                    new_links.append(link)
            param.links = new_links

    def _set_parameter_in_control_or_loose(self, set_param: prof.SetParameter) -> None:
        """
        Find the control with the param_id in it and set the parameter contents.

        It modifies controls in the control_dict not the catalog.
        Parameters are either bound to a control or are 'loose' and bound to the catalog itself.
        """
        # find the target param in control or the catalog's loose ones, i.e. catalog.params
        control = self._catalog_interface.get_control_by_param_id(set_param.param_id)
        loose_param = False
        if control:
            control.params = as_list(control.params)
            param_ids = [param.id for param in control.params]
            if set_param.param_id not in param_ids:
                raise TrestleNotFoundError(f'Param id {set_param.param_id} not found in control {control.id}')
            index = param_ids.index(set_param.param_id)
            param = control.params[index]
        else:
            param = self._catalog_interface.loose_param_dict.get(set_param.param_id, None)
            if param:
                loose_param = True
            else:
                logger.warning(f'SetParameter for param_id {set_param.param_id} not found in catalog')
                return

        # rules here follow https://pages.nist.gov/OSCAL/concepts/processing/profile-resolution/
        # see 'Modify Phase' and Setting Parameters

        Modify._set_overwrite_items(param, set_param)
        Modify._set_appended_items(param, set_param)
        Modify._set_replaced_or_appended_items(param, set_param)

        if loose_param:
            self._catalog_interface.loose_param_dict[set_param.param_id] = param
        else:
            control.params[index] = param
            self._catalog_interface.replace_control(control)

    def _modify_controls(self, catalog: cat.Catalog) -> cat.Catalog:
        """Modify the controls based on the profile."""
        logger.debug(f'modify specify catalog {catalog.metadata.title} for profile {self._profile.metadata.title}')
        self._catalog_interface = CatalogInterface(catalog)
        alters: Optional[List[prof.Alter]] = None
        # find the modify and alters
        if self._profile.modify:
            # change all parameter values
            if self._profile.modify.set_parameters and not self._block_params:
                set_param_list = self._profile.modify.set_parameters
                for set_param in set_param_list:
                    self._set_parameter_in_control_or_loose(set_param)
            alters = self._profile.modify.alters

        # an add with no by-id applies to the control
        # if position is starting it should appear immediately after title
        # if position is ending it should appear at end
        # if not specified it defaults to ending
        # if no by-id then before is treated as starting and after is treated as ending
        if alters is not None:
            title = self._profile.metadata.title
            for alter in alters:
                if alter.control_id is None:
                    logger.warning(f'Alter must have control id specified in profile {title}.')
                    continue
                id_ = alter.control_id
                if alter.removes is not None:
                    logger.warning(f'Alter not supported for removes in profile {title} control {id_}')
                    continue
                # we want a warning about adds even if adds are blocked, as in profile generate
                if alter.adds is None:
                    logger.warning(f'Alter has no adds in profile {title} control {id_}')
                    continue
                if not self._block_adds:
                    for add in alter.adds:
                        control = self._catalog_interface.get_control(id_)
                        if control is None:
                            logger.warning(
                                f'Alter/Add refers to control {id_} but it is not found in the import '
                                + f'for profile {self._profile.metadata.title}'
                            )
                        else:
                            self._add_to_control(control, add)
                            self._catalog_interface.replace_control(control)

        if self._change_prose:
            # go through all controls and fix the prose based on param values
            self._catalog_interface._change_prose_with_param_values(
                self._params_format,
                self._param_rep,
                self.show_value_warnings,
                self._value_assigned_prefix,
                self._value_not_assigned_prefix
            )

        catalog = self._catalog_interface.get_catalog()

        # update the original profile metadata with new contents
        # roles and responsible-parties will be pulled in with new uuid's
        # the title simply becomes the title of the current profile - and will get overwritten
        # by any other profiles downstream so that only the final profile title is used
        new_metadata = self._profile.metadata
        new_metadata.title = self._profile.metadata.title
        new_metadata.oscal_version = OSCAL_VERSION

        links: List[common.Link] = []
        for import_ in self._profile.imports:
            links.append(common.Link(**{'href': import_.href, 'rel': RESOLUTION_SOURCE}))
        new_metadata.links = links

        # move catalog controls from dummy group '' into the catalog
        for group in as_list(catalog.groups):
            if not group.id:
                catalog.controls = group.controls
                catalog.groups.remove(group)
                break

        catalog.metadata = new_metadata

        return catalog

    def process(self, catalog_iter: Iterator[cat.Catalog]) -> Iterator[cat.Catalog]:
        """Make the modifications to the controls based on the profile."""
        catalog = next(catalog_iter)
        logger.debug(
            f'modify process with catalog {catalog.metadata.title} using profile {self._profile.metadata.title}'
        )
        yield self._modify_controls(catalog)



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create ssp from catalog and profile."""

import argparse
import logging
import os
import pathlib
from typing import Any, Dict, List, Optional, Set

from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError

import trestle.core.generators as gens
import trestle.core.generic_oscal as generic
import trestle.oscal.common as com
import trestle.oscal.component as comp
import trestle.oscal.profile as prof
import trestle.oscal.ssp as ossp
from trestle.common import const, file_utils, log
from trestle.common.err import TrestleError, handle_generic_command_exception
from trestle.common.list_utils import as_list, comma_sep_to_list, delete_list_from_list, none_if_empty
from trestle.common.load_validate import load_validate_model_name
from trestle.common.model_utils import ModelUtils
from trestle.core.catalog.catalog_api import CatalogAPI
from trestle.core.catalog.catalog_interface import CatalogInterface
from trestle.core.catalog.catalog_reader import CatalogReader
from trestle.core.commands.author.common import AuthorCommonCommand
from trestle.core.commands.author.component import ComponentAssemble
from trestle.core.commands.common.cmd_utils import clear_folder
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.control_context import ContextPurpose, ControlContext
from trestle.core.control_interface import ControlInterface, ParameterRep
from trestle.core.control_reader import ControlReader
from trestle.core.crm.ssp_inheritance_api import SSPInheritanceAPI
from trestle.core.models.file_content_type import FileContentType
from trestle.core.profile_resolver import ProfileResolver
from trestle.core.remote.cache import FetcherFactory
from trestle.core.validator import Validator
from trestle.core.validator_factory import validator_factory

logger = logging.getLogger(__name__)


class SSPGenerate(AuthorCommonCommand):
    """Generate SSP in markdown form from a Profile."""

    name = 'ssp-generate'

    def _init_arguments(self) -> None:
        file_help_str = 'Main profile href, or name of the profile model in the trestle workspace'
        self.add_argument('-p', '--profile', help=file_help_str, required=True, type=str)
        self.add_argument(
            '-o', '--output', help='Name of the output generated ssp markdown folder', required=True, type=str
        )  # noqa E501
        self.add_argument('-cd', '--compdefs', help=const.HELP_COMPDEFS, required=False, type=str)

        ls_help_str = 'Leveraged ssp with inheritable controls href or name in the trestle_workspace'
        self.add_argument('-ls', '--leveraged-ssp', help=ls_help_str, required=False, type=str)

        self.add_argument('-y', '--yaml-header', help=const.HELP_YAML_PATH, required=False, type=str)
        self.add_argument(
            '-fo', '--force-overwrite', help=const.HELP_FO_OUTPUT, required=False, action='store_true', default=False
        )
        self.add_argument(
            '-ohv',
            '--overwrite-header-values',
            help=const.HELP_OVERWRITE_HEADER_VALUES,
            required=False,
            action='store_true',
            default=False
        )

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root = args.trestle_root
            if not file_utils.is_directory_name_allowed(args.output):
                raise TrestleError(f'{args.output} is not an allowed directory name')

            yaml_header: Dict[str, Any] = {}
            if args.yaml_header:
                try:
                    logging.debug(f'Loading yaml header file {args.yaml_header}')
                    yaml = YAML()
                    yaml_header = yaml.load(pathlib.Path(args.yaml_header).open('r'))
                except YAMLError as e:
                    raise TrestleError(f'YAML error loading yaml header {args.yaml_header} for ssp generation: {e}')

            compdef_name_list = comma_sep_to_list(args.compdefs)

            md_path = trestle_root / args.output

            return self._generate_ssp_markdown(
                trestle_root,
                args.profile,
                compdef_name_list,
                args.leveraged_ssp,
                md_path,
                yaml_header,
                args.overwrite_header_values,
                args.force_overwrite
            )

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error while writing markdown from catalog')

    def _generate_ssp_markdown(
        self,
        trestle_root: pathlib.Path,
        profile_name_or_href: str,
        compdef_name_list: List[str],
        leveraged_ssp_name_or_href: str,
        md_path: pathlib.Path,
        yaml_header: Dict[str, Any],
        overwrite_header_values: bool,
        force_overwrite: bool
    ) -> int:
        """
        Generate the ssp markdown from the profile and compdefs.

        Notes:
        Get RPC from profile.
        For each compdef:
            For each comp:
                Load top level rules
                for each control_imp:
                    Load rules params values
                    For each imp_req (bound to 1 control):
                        Load control level rules and status
                        Load part level rules
                        If rules apply then write out control and add to list written out
                        If control exists, read it and insert content
        """
        if force_overwrite:
            try:
                logger.info(f'Overwriting the content in {md_path}.')
                clear_folder(pathlib.Path(md_path))
            except TrestleError as e:  # pragma: no cover
                raise TrestleError(f'Unable to overwrite contents of {md_path}: {e}')

        context = ControlContext.generate(ContextPurpose.SSP, True, trestle_root, md_path)
        context.cli_yaml_header = yaml_header
        context.sections_dict = None
        context.prompt_responses = True
        context.overwrite_header_values = overwrite_header_values
        context.allowed_sections = None
        context.comp_def_name_list = compdef_name_list

        # if file not recognized as URI form, assume it represents name of file in trestle directory
        profile_in_trestle_dir = '://' not in profile_name_or_href
        profile_href = profile_name_or_href
        if profile_in_trestle_dir:
            local_path = f'profiles/{profile_name_or_href}/profile.json'
            profile_href = const.TRESTLE_HREF_HEADING + local_path
            profile_path = trestle_root / local_path
            _, _, context.profile = ModelUtils.load_distributed(profile_path, trestle_root)
        else:
            fetcher = FetcherFactory.get_fetcher(trestle_root, profile_href)
            context.profile = fetcher.get_oscal()
            profile_path = profile_href

        profile_resolver = ProfileResolver()
        # in ssp context we want to see missing value warnings
        resolved_catalog = profile_resolver.get_resolved_profile_catalog(
            trestle_root,
            profile_path,
            block_params=False,
            params_format='[.]',
            param_rep=ParameterRep.ASSIGNMENT_FORM,
            show_value_warnings=True
        )

        catalog_api = CatalogAPI(catalog=resolved_catalog, context=context)

        context.cli_yaml_header[const.TRESTLE_GLOBAL_TAG] = {}
        profile_header = {'title': context.profile.metadata.title, 'href': profile_href}

        context.cli_yaml_header[const.TRESTLE_GLOBAL_TAG][const.PROFILE] = profile_header

        catalog_api.write_catalog_as_markdown()

        # Generate inheritance view after controls view completes
        if leveraged_ssp_name_or_href:
            self._generate_inheritance_markdown(trestle_root, leveraged_ssp_name_or_href, resolved_catalog, md_path)

        return CmdReturnCodes.SUCCESS.value

    def _generate_inheritance_markdown(
        self,
        trestle_root: pathlib.Path,
        leveraged_ssp_name_or_href: str,
        resolved_catalog: CatalogInterface,
        md_path: str
    ) -> None:
        """
        Generate markdown for inheritance view.

        Notes:
            This will create the inheritance view markdown files in the same directory as the ssp markdown files.
            The information will be from the leveraged ssp, but filtered by the chosen profile to ensure only relevant
            controls are present for mapping.
        """
        # if file not recognized as URI form, assume it represents name of file in trestle directory
        ssp_in_trestle_dir = '://' not in leveraged_ssp_name_or_href
        ssp_href = leveraged_ssp_name_or_href
        if ssp_in_trestle_dir:
            local_path = f'{const.MODEL_DIR_SSP}/{leveraged_ssp_name_or_href}/system-security-plan.json'
            ssp_href = const.TRESTLE_HREF_HEADING + local_path

        inheritance_view_path: pathlib.Path = md_path.joinpath(const.INHERITANCE_VIEW_DIR)
        inheritance_view_path.mkdir(exist_ok=True)
        logger.debug(f'Creating content for inheritance view in {inheritance_view_path}')

        ssp_inheritance_api = SSPInheritanceAPI(inheritance_view_path, trestle_root)

        # Filter the ssp implemented requirements by the catalog specified
        catalog_api: CatalogAPI = CatalogAPI(catalog=resolved_catalog)
        ssp_inheritance_api.write_inheritance_as_markdown(ssp_href, catalog_api)


class SSPAssemble(AuthorCommonCommand):
    """Assemble markdown files of controls into an SSP json file."""

    name = 'ssp-assemble'

    def _init_arguments(self) -> None:
        name_help_str = (
            'Optional name of the ssp model in the trestle workspace that is being modified.  '
            'If not provided the output name is used.'
        )
        self.add_argument('-n', '--name', help=name_help_str, required=False, type=str)
        file_help_str = 'Name of the input markdown file directory'
        self.add_argument('-m', '--markdown', help=file_help_str, required=True, type=str)
        output_help_str = 'Name of the output generated json SSP'
        self.add_argument('-cd', '--compdefs', help=const.HELP_COMPDEFS, required=False, type=str)
        self.add_argument('-o', '--output', help=output_help_str, required=True, type=str)
        self.add_argument('-r', '--regenerate', action='store_true', help=const.HELP_REGENERATE)
        self.add_argument('-vn', '--version', help=const.HELP_VERSION, required=False, type=str)

    @staticmethod
    def _get_ssp_component(ssp: ossp.SystemSecurityPlan, gen_comp: generic.GenericComponent) -> ossp.SystemComponent:
        for component in as_list(ssp.system_implementation.components):
            if component.title == gen_comp.title:
                return component
        # if this is a new system component assign its status as operational by default
        # the status of the system components are not stored in the markdown
        gen_comp.status.state = const.STATUS_OPERATIONAL
        new_component = gen_comp.as_system_component()
        return new_component

    @staticmethod
    def _merge_by_comps(stat: ossp.Statement, statement: ossp.Statement, set_params: List[ossp.SetParameter]) -> None:
        for by_comp in as_list(statement.by_components):
            found = False
            for dest_by_comp in as_list(stat.by_components):
                if dest_by_comp.component_uuid == by_comp.component_uuid:
                    dest_by_comp.description = by_comp.description
                    dest_by_comp.props = as_list(dest_by_comp.props)
                    dest_by_comp.props.extend(as_list(statement.props))
                    dest_by_comp.props = none_if_empty(ControlInterface.clean_props(by_comp.props))
                    dest_by_comp.implementation_status = by_comp.implementation_status
                    dest_by_comp.set_parameters = none_if_empty(set_params)
                    found = True
                    break
            if not found:
                stat.by_components = as_list(stat.by_components)
                by_comp.set_parameters = none_if_empty(set_params)
                stat.by_components.append(by_comp)

    @staticmethod
    def _merge_statement(
        imp_req: ossp.ImplementedRequirement,
        statement: generic.GenericStatement,
        set_params: List[ossp.SetParameter],
    ) -> None:
        """Merge the generic statement into the statements of the imp_req."""
        # if the statement id is already in the imp_req, merge its by_comps into the existing statement
        for stat in as_list(imp_req.statements):
            if stat.statement_id == statement.statement_id:
                SSPAssemble._merge_by_comps(stat, statement, set_params)
                return
        # otherwise just ad the statement - but only if it has by_comps
        if statement.by_components:
            imp_req.statements = as_list(imp_req.statements)
            imp_req.statements.append(statement)

    @staticmethod
    def _merge_imp_req_into_imp_req(
        imp_req: ossp.ImplementedRequirement,
        gen_imp_req: generic.GenericImplementedRequirement,
        set_params: List[ossp.SetParameter]
    ) -> None:
        """Merge comp def imp req into existing imp req."""
        # convert generic imp req from comp defs into ssp form
        src_imp_req = gen_imp_req.as_ssp()
        imp_req.props = none_if_empty(
            ControlInterface.clean_props(gen_imp_req.props, remove_imp_status=True, remove_all_rule_info=True)
        )
        for statement in as_list(src_imp_req.statements):
            SSPAssemble._merge_statement(imp_req, statement, set_params)

    @staticmethod
    def _get_params_for_rules(context: ControlContext, rules_list: List[str],
                              set_params: List[ossp.SetParameter]) -> List[ossp.SetParameter]:
        """Get all set_params needed by the rules along with non-rule set_params."""
        needed_param_ids: Set[str] = set()
        rule_dict = context.rules_params_dict.get(context.comp_name, {})
        # find param_ids needed by rules
        for rule_id in rules_list:
            # get list of param_ids associated with this rule_id
            param_ids = [
                param['name'] for params in rule_dict.values() for param in params if param['rule-id'] == rule_id
            ]
            needed_param_ids.update(param_ids)
        all_rule_param_ids = [param['name'] for params in rule_dict.values() for param in params]
        # any set_param that isn't associated with a rule should be included as a normal control set param with no rule
        for set_param in set_params:
            if set_param.param_id not in all_rule_param_ids:
                needed_param_ids.add(set_param.param_id)
        param_ids_list = sorted(needed_param_ids)
        needed_set_params: List[ossp.SetParameter] = []
        for param_id in param_ids_list:
            set_param = None
            for sp in set_params:
                if sp.param_id == param_id:
                    set_param = sp
                    break
            if set_param:
                needed_set_params.append(set_param)
            else:
                logger.warning(f'No set param found for param {param_id}')
        return needed_set_params

    @staticmethod
    def _add_imp_req_to_ssp(
        ssp: ossp.SystemSecurityPlan,
        gen_comp: generic.GenericComponent,
        gen_imp_req: generic.GenericImplementedRequirement,
        set_params: List[ossp.SetParameter],
        context: ControlContext
    ) -> None:
        """Add imp req from control implementation into new ssp being assembled."""
        # the incoming gen_imp_req comes directly from the comp def
        # but the imp_req here is pulled from the ssp and created if not already there
        imp_req = CatalogReader._get_imp_req_for_control(ssp, gen_imp_req.control_id)
        local_set_params = as_list(set_params)[:]
        local_set_params.extend(as_list(imp_req.set_parameters))
        local_set_params = ControlInterface.uniquify_set_params(local_set_params)
        # get any rules set at control level, if present
        rules_list, _ = ControlInterface.get_rule_list_for_item(gen_imp_req)  # type: ignore
        # There should be no rule content at top level of imp_req in ssp so strip them out
        imp_req.props = none_if_empty(
            ControlInterface.clean_props(gen_imp_req.props, remove_imp_status=True, remove_all_rule_info=True)
        )
        # if we have rules applying or need to make set_params, we need to make a by_comp
        control_set_params = SSPAssemble._get_params_for_rules(context, rules_list, local_set_params)
        if rules_list or control_set_params:
            by_comp = gens.generate_sample_model(ossp.ByComponent)
            by_comp.component_uuid = gen_comp.uuid
            by_comp.description = gen_imp_req.description
            by_comp.set_parameters = none_if_empty(control_set_params)
            by_comp.implementation_status = ControlInterface.get_status_from_props(gen_imp_req)  # type: ignore
            by_comp.props = none_if_empty(ControlInterface.clean_props(gen_imp_req.props))
            imp_req.by_components = as_list(imp_req.by_components)
            imp_req.by_components.append(by_comp)
        # each statement in ci corresponds to by_comp in an ssp imp req
        # so insert the new by_comp directly into the ssp, generating parts as needed
        imp_req.statements = as_list(imp_req.statements)
        for statement in as_list(gen_imp_req.statements):
            if ControlInterface.item_has_rules(statement):  # type: ignore
                imp_req = CatalogReader._get_imp_req_for_statement(ssp, gen_imp_req.control_id, statement.statement_id)
                by_comp = CatalogReader._get_by_comp_from_imp_req(imp_req, statement.statement_id, gen_comp.uuid)
                by_comp.description = statement.description
                by_comp.props = none_if_empty(ControlInterface.clean_props(statement.props))
                rules_list, _ = ControlInterface.get_rule_list_for_item(statement)  # type: ignore
                by_comp.set_parameters = none_if_empty(
                    SSPAssemble._get_params_for_rules(context, rules_list, local_set_params)
                )
        imp_req.statements = none_if_empty(imp_req.statements)
        ssp.control_implementation.implemented_requirements = as_list(
            ssp.control_implementation.implemented_requirements
        )

    @staticmethod
    def _merge_imp_req_into_ssp(
        ssp: ossp.SystemSecurityPlan,
        gen_imp_req: generic.GenericImplementedRequirement,
        set_params: List[ossp.SetParameter],
    ) -> None:
        """Merge the new imp_reqs into the ssp."""
        for imp_req in as_list(ssp.control_implementation.implemented_requirements):
            if imp_req.control_id == gen_imp_req.control_id:
                SSPAssemble._merge_imp_req_into_imp_req(imp_req, gen_imp_req, set_params)
                return
        new_imp_req = gen_imp_req.as_ssp()
        imp_req.props = none_if_empty(
            ControlInterface.clean_props(gen_imp_req.props, remove_imp_status=True, remove_all_rule_info=True)
        )
        ssp.control_implementation.implemented_requirements.append(new_imp_req)

    def _merge_comp_defs(
        self,
        ssp: ossp.SystemSecurityPlan,
        comp_dict: Dict[str, generic.GenericComponent],
        context: ControlContext,
        catalog_interface: CatalogInterface
    ) -> None:
        """Merge the original generic comp defs into the ssp."""
        all_comps: List[ossp.SystemComponent] = []
        # determine if this is a new and empty ssp
        new_ssp = not ssp.control_implementation.implemented_requirements
        for _, gen_comp in comp_dict.items():
            context.comp_name = gen_comp.title
            all_ci_props: List[com.Property] = []
            ssp_comp = SSPAssemble._get_ssp_component(ssp, gen_comp)
            set_params: List[ossp.SetParameter] = []
            for ci in as_list(gen_comp.control_implementations):
                all_ci_props.extend(as_list(ci.props))
                # get the list of set_params in the control implementation - for this component
                for sp in as_list(ci.set_parameters):
                    set_params.append(sp.to_ssp())
                for imp_req in as_list(ci.implemented_requirements):
                    # ignore any controls not in the ssp profile (resolved catalog)
                    if not catalog_interface.get_control(imp_req.control_id):
                        logger.debug(f'Ignoring imp_req for control {imp_req.control_id} not in ssp profile')
                        continue
                    if new_ssp:
                        SSPAssemble._add_imp_req_to_ssp(ssp, gen_comp, imp_req, set_params, context)
                    else:
                        # compile all new uuids for new component definitions
                        comp_uuids = [x.uuid for x in comp_dict.values()]
                        for imp_requirement in as_list(ssp.control_implementation.implemented_requirements):
                            to_delete = []
                            for i, by_comp in enumerate(imp_requirement.by_components):
                                if by_comp.component_uuid not in comp_uuids:
                                    logger.warning(
                                        f'By_component {by_comp.component_uuid} removed from implemented requirement '
                                        f'{imp_requirement.control_id} because the corresponding component is not in '
                                        'the specified compdefs '
                                    )
                                    to_delete.append(i)
                            if to_delete:
                                delete_list_from_list(imp_requirement.by_components, to_delete)
                        SSPAssemble._merge_imp_req_into_ssp(ssp, imp_req, set_params)
            ssp_comp.props = as_list(gen_comp.props)
            ssp_comp.props.extend(all_ci_props)
            ssp_comp.props = none_if_empty(ControlInterface.clean_props(ssp_comp.props))
            all_comps.append(ssp_comp)

        ssp.system_implementation.components = none_if_empty(all_comps)

    def _generate_roles_in_metadata(self, ssp: ossp.SystemSecurityPlan) -> bool:
        """Find all roles referenced by imp reqs and create role in metadata as needed."""
        metadata = ssp.metadata
        metadata.roles = as_list(metadata.roles)
        known_role_ids = [role.id for role in metadata.roles]
        changed = False
        for imp_req in ssp.control_implementation.implemented_requirements:
            role_ids = [resp_role.role_id for resp_role in as_list(imp_req.responsible_roles)]
            for role_id in role_ids:
                if role_id not in known_role_ids:
                    role = com.Role(id=role_id, title=role_id)
                    metadata.roles.append(role)
                    known_role_ids.append(role_id)
                    changed = True
        metadata.roles = none_if_empty(metadata.roles)
        return changed

    @staticmethod
    def _build_comp_dict_from_comp_defs(
        trestle_root: pathlib.Path, comp_def_name_list: List[str], create_sys_comp: bool
    ) -> Dict[str, generic.GenericComponent]:
        comp_dict: Dict[str, generic.GenericComponent] = {}
        for comp_name in comp_def_name_list:
            comp_def, _ = ModelUtils.load_model_for_class(trestle_root, comp_name, comp.ComponentDefinition)
            for def_comp in as_list(comp_def.components):
                gen_def_comp = generic.GenericComponent.from_defined_component(def_comp)
                comp_dict[def_comp.title] = gen_def_comp
        if create_sys_comp:
            sys_comp = generic.GenericComponent.generate()
            sys_comp.type = const.THIS_SYSTEM_AS_KEY
            sys_comp.title = const.SSP_MAIN_COMP_NAME
            comp_dict[sys_comp.title] = sys_comp
        return comp_dict

    @staticmethod
    def _get_this_system_as_gen_comp(ssp: ossp.SystemSecurityPlan) -> Optional[generic.GenericComponent]:
        for component in as_list(ssp.system_implementation.components):
            if component.title == const.SSP_MAIN_COMP_NAME:
                return generic.GenericComponent.from_defined_component(component)
        return None

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root = pathlib.Path(args.trestle_root)

            md_path = trestle_root / args.markdown

            # the original, reference ssp name defaults to same as output if name not specified
            # thus in cyclic editing you are reading and writing same json ssp
            orig_ssp_name = args.output
            if args.name:
                orig_ssp_name = args.name
            new_ssp_name = args.output

            _, profile_href = ComponentAssemble._get_profile_title_and_href_from_dir(md_path)
            res_cat = ProfileResolver.get_resolved_profile_catalog(
                trestle_root, profile_href, param_rep=ParameterRep.LEAVE_MOUSTACHE
            )
            catalog_interface = CatalogInterface(res_cat)

            new_file_content_type = FileContentType.JSON

            # if output ssp already exists, load it to see if new one is different
            existing_ssp: Optional[ossp.SystemSecurityPlan] = None
            new_ssp_path = ModelUtils.get_model_path_for_name_and_class(
                trestle_root, new_ssp_name, ossp.SystemSecurityPlan
            )
            if new_ssp_path:
                _, _, existing_ssp = ModelUtils.load_distributed(new_ssp_path, trestle_root)
                new_file_content_type = FileContentType.path_to_content_type(new_ssp_path)

            ssp: ossp.SystemSecurityPlan

            # if orig ssp exists - need to load it rather than instantiate new one
            orig_ssp_path = ModelUtils.get_model_path_for_name_and_class(
                trestle_root, orig_ssp_name, ossp.SystemSecurityPlan
            )

            context = ControlContext.generate(ContextPurpose.SSP, True, trestle_root, md_path)
            context.comp_def_name_list = comma_sep_to_list(args.compdefs)
            part_id_map_by_id = catalog_interface.get_statement_part_id_map(False)
            catalog_interface.generate_control_rule_info(part_id_map_by_id, context)

            # load all original comp defs
            # only additions from markdown will be imp_req prose and status
            # and param vals
            # if this is a new ssp then create system component in the comp_dict
            comp_dict = SSPAssemble._build_comp_dict_from_comp_defs(
                trestle_root, context.comp_def_name_list, not orig_ssp_path
            )

            part_id_map_by_label = catalog_interface.get_statement_part_id_map(True)

            # if ssp already exists use it as container for new content
            if orig_ssp_path:
                # load the existing json ssp
                _, _, ssp = ModelUtils.load_distributed(orig_ssp_path, trestle_root)
                # add the This System comp to the comp dict so its uuid is known
                sys_comp = SSPAssemble._get_this_system_as_gen_comp(ssp)
                if not sys_comp:
                    raise TrestleError('Original ssp has no system component.')
                comp_dict[const.SSP_MAIN_COMP_NAME] = sys_comp

                ssp_sys_imp_comps = ssp.system_implementation.components
                # Gather the leveraged components to add back after the merge
                leveraged_comps: Dict[str, ossp.SystemComponent] = {}
                for sys_comp in ssp_sys_imp_comps:
                    if sys_comp.props is not None:
                        prop_names = [x.name for x in sys_comp.props]
                        if const.LEV_AUTH_UUID in prop_names:
                            leveraged_comps[sys_comp.title] = sys_comp

                # Verifies older compdefs in an ssp no longer exist in newly provided ones
                comp_titles = [x.title for x in comp_dict.values()]
                diffs = [x for x in ssp_sys_imp_comps if x.title not in comp_titles and x.title not in leveraged_comps]
                if diffs:
                    for diff in diffs:
                        logger.warning(
                            f'Component named: {diff.title} was removed from system components from ssp '
                            'because the corresponding component is not in '
                            'the specified compdefs '
                        )
                    index_list = [ssp_sys_imp_comps.index(value) for value in diffs if value in ssp_sys_imp_comps]
                    delete_list_from_list(ssp.system_implementation.components, index_list)

                self._merge_comp_defs(ssp, comp_dict, context, catalog_interface)
                CatalogReader.read_ssp_md_content(md_path, ssp, comp_dict, part_id_map_by_label, context)

                new_file_content_type = FileContentType.path_to_content_type(orig_ssp_path)

                # Add the leveraged comps back to the final ssp
                ssp.system_implementation.components.extend(list(leveraged_comps.values()))
            else:
                # create a sample ssp to hold all the parts
                ssp = gens.generate_sample_model(ossp.SystemSecurityPlan)
                ssp.control_implementation.implemented_requirements = []
                ssp.control_implementation.description = const.SSP_SYSTEM_CONTROL_IMPLEMENTATION_TEXT
                ssp.system_implementation.components = []
                self._merge_comp_defs(ssp, comp_dict, context, catalog_interface)
                CatalogReader.read_ssp_md_content(md_path, ssp, comp_dict, part_id_map_by_label, context)

                import_profile: ossp.ImportProfile = gens.generate_sample_model(ossp.ImportProfile)
                import_profile.href = const.REPLACE_ME
                ssp.import_profile = import_profile

            # now that we know the complete list of needed components, add them to the sys_imp
            # TODO if the ssp already existed then components may need to be removed if not ref'd by imp_reqs
            self._generate_roles_in_metadata(ssp)

            # If this is a leveraging SSP, update it with the retrieved exports from the leveraged SSP
            inheritance_markdown_path = md_path.joinpath(const.INHERITANCE_VIEW_DIR)
            if os.path.exists(inheritance_markdown_path):
                SSPInheritanceAPI(inheritance_markdown_path, trestle_root).update_ssp_inheritance(ssp)

            ssp.import_profile.href = profile_href

            if args.version:
                ssp.metadata.version = args.version

            if ModelUtils.models_are_equivalent(existing_ssp, ssp):
                logger.info('No changes to assembled ssp so ssp not written out.')
                return CmdReturnCodes.SUCCESS.value

            if args.regenerate:
                ssp, _, _ = ModelUtils.regenerate_uuids(ssp)
            ModelUtils.update_last_modified(ssp)
            # validate model rules before saving
            args_validate = argparse.Namespace(mode=const.VAL_MODE_RULES)
            validator: Validator = validator_factory.get(args_validate)
            if not validator.model_is_valid(ssp, True, trestle_root):
                logger.error(
                    'Validation of file to be imported did not pass. Rule parameter values validation failed, '
                    'please check values are correct for shared parameters in current model'
                )
                return CmdReturnCodes.COMMAND_ERROR.value
            # write out the ssp as json
            ModelUtils.save_top_level_model(ssp, trestle_root, new_ssp_name, new_file_content_type)

            return CmdReturnCodes.SUCCESS.value

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error while assembling SSP')


class SSPFilter(AuthorCommonCommand):
    """
    Filter the controls in an ssp.

    The filtered ssp is based on controls included by the following:
    profile, components, implementation status, and/or control origination.
    """

    name = 'ssp-filter'

    def _init_arguments(self) -> None:
        file_help_str = 'Name of the input ssp'
        self.add_argument('-n', '--name', help=file_help_str, required=True, type=str)
        file_help_str = 'Name of the optional input profile that defines set of controls in filtered ssp'
        self.add_argument('-p', '--profile', help=file_help_str, required=False, type=str)
        output_help_str = 'Name of the output generated SSP'
        self.add_argument('-o', '--output', help=output_help_str, required=True, type=str)
        self.add_argument('-r', '--regenerate', action='store_true', help=const.HELP_REGENERATE)
        self.add_argument('-vn', '--version', help=const.HELP_VERSION, required=False, type=str)
        comp_help_str = 'Colon-delimited list of component names to include in filtered ssp.'
        self.add_argument('-c', '--components', help=comp_help_str, required=False, type=str)
        is_help_str = 'Comma-delimited list of control implementation statuses to include in filtered ssp.'
        self.add_argument('-is', '--implementation-status', help=is_help_str, required=False, type=str)
        co_help_str = 'Comma-delimited list of control origination values to include in filtered ssp.'
        self.add_argument('-co', '--control-origination', help=co_help_str, required=False, type=str)

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root = pathlib.Path(args.trestle_root)
            comp_names: Optional[List[str]] = None
            impl_status_values: Optional[List[str]] = None
            co_values: Optional[List[str]] = None

            if not (args.components or args.implementation_status or args.profile or args.control_origination):
                logger.warning(
                    'You must specify at least one, or a combination of: profile, list of component names'
                    ', list of implementation statuses, or list of control origination values for ssp-filter.'
                )
                return CmdReturnCodes.COMMAND_ERROR.value

            if args.components:
                comp_names = args.components.split(':')

            if args.implementation_status:
                impl_status_values = args.implementation_status.split(',')
                allowed_is_values = {
                    const.STATUS_PLANNED,
                    const.STATUS_PARTIAL,
                    const.STATUS_IMPLEMENTED,
                    const.STATUS_ALTERNATIVE,
                    const.STATUS_NOT_APPLICABLE
                }
                allowed_is_string = ', '.join(str(item) for item in allowed_is_values)
                for impl_status in impl_status_values:
                    if impl_status not in allowed_is_values:
                        logger.warning(
                            f'Provided implementation status "{impl_status}" is invalid.\n'
                            f'Please use the following for ssp-filter: {allowed_is_string}'
                        )
                        return CmdReturnCodes.COMMAND_ERROR.value

            if args.control_origination:
                co_values = args.control_origination.split(',')
                allowed_co_values = {
                    const.ORIGINATION_ORGANIZATION,
                    const.ORIGINATION_SYSTEM_SPECIFIC,
                    const.ORIGINATION_INHERITED,
                    const.ORIGINATION_CUSTOMER_CONFIGURED,
                    const.ORIGINATION_CUSTOMER_PROVIDED
                }
                allowed_co_string = ', '.join(str(item) for item in allowed_co_values)
                for co in co_values:
                    if co not in allowed_co_values:
                        logger.warning(
                            f'Provided control origination "{co}" is invalid.\n'
                            f'Please use the following for ssp-filter: {allowed_co_string}'
                        )
                        return CmdReturnCodes.COMMAND_ERROR.value

            return self.filter_ssp(
                trestle_root,
                args.name,
                args.profile,
                args.output,
                args.regenerate,
                args.version,
                comp_names,
                impl_status_values,
                co_values
            )
        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error generating the filtered ssp')

    def filter_ssp(
        self,
        trestle_root: pathlib.Path,
        ssp_name: str,
        profile_name: str,
        out_name: str,
        regenerate: bool,
        version: Optional[str],
        components: Optional[List[str]] = None,
        implementation_status: Optional[List[str]] = None,
        control_origination: Optional[List[str]] = None
    ) -> int:
        """
        Filter the ssp and output new ssp.

        The filtered ssp is based on controls included by the following:
        profile, components, implementation status, and/or control origination.

        Args:
            trestle_root: root directory of the trestle workspace
            ssp_name: name of the ssp model
            profile_name: name of the optional profile model used for filtering
            out_name: name of the output ssp model with filtered controls
            regenerate: whether to regenerate the uuid's in the ssp
            version: new version for the model
            components: optional list of component names used for filtering
            implementation_status: optional list of implementation statuses for filtering
            control_origination: optional list of control origination values for filtering

        Returns:
            0 on success, 1 otherwise
        """
        # load the ssp
        ssp: ossp.SystemSecurityPlan
        ssp, _ = load_validate_model_name(trestle_root, ssp_name, ossp.SystemSecurityPlan, FileContentType.JSON)
        profile_path = ModelUtils.get_model_path_for_name_and_class(
            trestle_root, profile_name, prof.Profile, FileContentType.JSON
        )

        if components:
            raw_comp_names = [ControlReader.simplify_name(name) for name in components]
            comp_uuids: List[str] = []
            for component in ssp.system_implementation.components:
                if ControlReader.simplify_name(component.title) in raw_comp_names:
                    comp_uuids.append(component.uuid)

            if len(comp_uuids) != len(components):
                raise TrestleError(
                    f'Unable to filter the ssp because one of the components {components} is not in the ssp.'
                )

            # imp_reqs can be by comp
            # and imp_reqs can have statements that are by comp
            if comp_uuids:
                new_imp_reqs: List[ossp.ImplementedRequirement] = []
                # these are all required to be present
                for imp_req in ssp.control_implementation.implemented_requirements:
                    new_by_comps: List[ossp.ByComponent] = []
                    # by_comps is optional
                    for by_comp in as_list(imp_req.by_components):
                        if by_comp.component_uuid in comp_uuids:
                            new_by_comps.append(by_comp)
                    imp_req.by_components = none_if_empty(new_by_comps)
                    new_imp_reqs.append(imp_req)
                    new_statements: List[ossp.Statement] = []
                    for statement in as_list(imp_req.statements):
                        new_by_comps: List[ossp.ByComponent] = []
                        for by_comp in as_list(statement.by_components):
                            if by_comp.component_uuid in comp_uuids:
                                new_by_comps.append(by_comp)
                        statement.by_components = none_if_empty(new_by_comps)
                        new_statements.append(statement)
                    imp_req.statements = none_if_empty(new_statements)
                ssp.control_implementation.implemented_requirements = new_imp_reqs
                # now remove any unused components from the ssp
                new_comp_list: List[ossp.SystemComponent] = []
                for comp_ in ssp.system_implementation.components:
                    if comp_.uuid in comp_uuids:
                        new_comp_list.append(comp_)
                ssp.system_implementation.components = new_comp_list

        # filter by controls in profile
        if profile_name:
            prof_resolver = ProfileResolver()
            catalog = prof_resolver.get_resolved_profile_catalog(trestle_root, profile_path, show_value_warnings=True)
            catalog_api = CatalogAPI(catalog=catalog)

            # The input ssp should reference a superset of the controls referenced by the profile
            # Need to cull references in the ssp to controls not in the profile
            # Also make sure the output ssp contains imp reqs for all controls in the profile
            control_imp = ssp.control_implementation
            ssp_control_ids: Set[str] = set()

            new_set_params: List[ossp.SetParameter] = []
            for set_param in as_list(control_imp.set_parameters):
                control = catalog_api._catalog_interface.get_control_by_param_id(set_param.param_id)
                if control is not None:
                    new_set_params.append(set_param)
                    ssp_control_ids.add(control.id)
            control_imp.set_parameters = none_if_empty(new_set_params)

            new_imp_requirements: List[ossp.ImplementedRequirement] = []
            for imp_requirement in as_list(control_imp.implemented_requirements):
                control = catalog_api._catalog_interface.get_control(imp_requirement.control_id)
                if control is not None:
                    new_imp_requirements.append(imp_requirement)
                    ssp_control_ids.add(control.id)
            control_imp.implemented_requirements = new_imp_requirements

            # make sure all controls in the profile have implemented reqs in the final ssp
            if not ssp_control_ids.issuperset(catalog_api._catalog_interface.get_control_ids()):
                raise TrestleError('Unable to filter the ssp because the profile references controls not in it.')

            ssp.control_implementation = control_imp

        # filter implemented requirements and statements by component implementation status
        # this will remove any implemented requirements without statements or by_component fields set
        if implementation_status:
            new_imp_reqs: List[ossp.ImplementedRequirement] = []
            # these are all required to be present
            for imp_req in ssp.control_implementation.implemented_requirements:
                new_by_comps: List[ossp.ByComponent] = []
                # by_comps is optional
                for by_comp in as_list(imp_req.by_components):
                    if by_comp.implementation_status.state in implementation_status:
                        new_by_comps.append(by_comp)
                imp_req.by_components = none_if_empty(new_by_comps)

                new_statements: List[ossp.Statement] = []
                for statement in as_list(imp_req.statements):
                    new_by_comps: List[ossp.ByComponent] = []
                    for by_comp in as_list(statement.by_components):
                        if by_comp.implementation_status.state in implementation_status:
                            new_by_comps.append(by_comp)
                    statement.by_components = none_if_empty(new_by_comps)
                    if statement.by_components is not None:
                        new_statements.append(statement)
                imp_req.statements = none_if_empty(new_statements)

                if imp_req.by_components is not None or imp_req.statements is not None:
                    new_imp_reqs.append(imp_req)

            ssp.control_implementation.implemented_requirements = new_imp_reqs

        # filter implemented requirements by control origination property.
        # this will remove any implemented requirements without the control origination
        # property set
        if control_origination:
            new_imp_reqs: List[ossp.ImplementedRequirement] = []

            for imp_requirement in ssp.control_implementation.implemented_requirements:
                if imp_requirement.props:
                    for prop in imp_requirement.props:
                        if prop.name == const.CONTROL_ORIGINATION and prop.value in control_origination:
                            new_imp_reqs.append(imp_requirement)
                            # only add the imp requirement one time
                            break

            ssp.control_implementation.implemented_requirements = new_imp_reqs

        if version:
            ssp.metadata.version = version

        existing_ssp_path = ModelUtils.get_model_path_for_name_and_class(
            trestle_root, out_name, ossp.SystemSecurityPlan
        )
        if existing_ssp_path is not None:
            existing_ssp, _ = load_validate_model_name(trestle_root, out_name, ossp.SystemSecurityPlan)
            if ModelUtils.models_are_equivalent(existing_ssp, ssp):  # type: ignore
                logger.info('No changes to filtered ssp so ssp not written out.')
                return CmdReturnCodes.SUCCESS.value

        if regenerate:
            ssp, _, _ = ModelUtils.regenerate_uuids(ssp)

        ModelUtils.update_last_modified(ssp)

        ModelUtils.save_top_level_model(ssp, trestle_root, out_name, FileContentType.JSON)

        return CmdReturnCodes.SUCCESS.value



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle describe command."""
import argparse
import os
import pathlib

import pytest

from tests import test_utils

from trestle.core.commands.describe import DescribeCmd
from trestle.core.models.file_content_type import FileContentType
from trestle.oscal import catalog as oscatalog
from trestle.oscal.component import ComponentDefinition


@pytest.mark.parametrize('element_path', ['', 'catalog.metadata.roles', 'catalog.metadata'])
def test_describe_functionality(
    element_path: str, tmp_path: pathlib.Path, keep_cwd: pathlib.Path, simplified_nist_catalog: oscatalog.Catalog
) -> None:
    """Test basic functionality of describe."""
    # prepare trestle project dir with the file
    catalog_dir, catalog_file = test_utils.prepare_trestle_project_dir(
        tmp_path,
        FileContentType.JSON,
        simplified_nist_catalog,
        test_utils.CATALOGS_DIR)

    os.chdir(catalog_dir)
    args = argparse.Namespace(file='catalog.json', element=element_path, verbose=1, trestle_root=tmp_path)
    assert DescribeCmd()._run(args) == 0

    results = DescribeCmd.describe(catalog_file.resolve(), '', tmp_path)
    assert len(results) > 5
    assert 'catalog.json' in results[0]
    assert 'catalog.Catalog' in results[0]


def test_describe_failures(
    tmp_path: pathlib.Path, keep_cwd: pathlib.Path, sample_component_definition: ComponentDefinition
) -> None:
    """Test describe failure modes."""
    comp_def_dir, comp_def_file = test_utils.prepare_trestle_project_dir(
        tmp_path,
        FileContentType.JSON,
        sample_component_definition,
        test_utils.COMPONENT_DEF_DIR)

    # not in trestle directory
    args = argparse.Namespace(
        file='comp_def.json', element='component-definition.metadata', verbose=1, trestle_root=tmp_path
    )
    assert DescribeCmd()._run(args) == 1

    os.chdir(comp_def_dir)

    # in trestle directory but have wildcard in element path
    args = argparse.Namespace(
        file=comp_def_file, element='component-definition.*.roles', verbose=1, trestle_root=tmp_path
    )
    assert DescribeCmd()._run(args) == 1

    # in trestle directory but have comma in element path
    args = argparse.Namespace(
        file=comp_def_file,
        element='component-definition.metadata,component-definition.back-matter',
        verbose=1,
        trestle_root=tmp_path
    )
    assert DescribeCmd()._run(args) == 1

    # in trestle directory but element only has one part
    args = argparse.Namespace(file=comp_def_file, element='component-definition', verbose=1, trestle_root=tmp_path)
    assert DescribeCmd()._run(args) == 1

    # no filename specified
    args = argparse.Namespace(file=None, element=None, trestle_root=tmp_path, verbose=1)
    assert DescribeCmd()._run(args) == 2



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle command abstraction.

Improves parsing until such a point as ILCLI is fixed.
"""

import argparse
import logging
from typing import Optional, TextIO

from ilcli import Command

from trestle.common import file_utils
from trestle.core.commands.common.return_codes import CmdReturnCodes

logger = logging.getLogger(__name__)


class CommandBase(Command):
    """Linear extension to the ILCLI interface to use documentation string more.

    Trestle commands not requiring trestle-root should extend from this class.
    """

    # Example commands extedning from this class - init', 'trestle', 'version', 'partial-object-validate'
    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
        parent: Optional[Command] = None,
        name: Optional[str] = None,
        out: Optional[TextIO] = None,
        err: Optional[TextIO] = None
    ) -> None:
        """Override default ILCLI behaviour to include class documentation in command help description."""
        super(CommandBase, self).__init__(parser, parent, name, out, err)
        self.parser.description = self.__doc__


class CommandPlusDocs(CommandBase):
    """This class validates trestle-root argument.

    Trestle commands requiring trestle-root should extend from this class.
    All commands that extend this class will validate the state of trestle workspace.
    """

    def _validate_arguments(self, args: argparse.ArgumentParser) -> int:
        """Check trestle-root argument is a valid trestle root directory."""
        root = file_utils.extract_trestle_project_root(args.trestle_root)  # type: ignore
        if root is None:
            logger.error(f'Given directory {args.trestle_root} is not in a valid trestle root directory')
            return CmdReturnCodes.TRESTLE_ROOT_ERROR.value
        is_oscal_dir_valid = file_utils.check_oscal_directories(args.trestle_root)  # type: ignore
        if not is_oscal_dir_valid:
            return CmdReturnCodes.TRESTLE_ROOT_ERROR.value
        args.trestle_root = root  # type: ignore
        return CmdReturnCodes.SUCCESS.value



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Create resolved catalog from profile."""

import logging
from typing import Iterator, List, Optional, Tuple, Union

import trestle.oscal.catalog as cat
import trestle.oscal.common as com
import trestle.oscal.profile as prof
from trestle.common.common_types import OBT
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_list, none_if_empty
from trestle.core.pipeline import Pipeline

logger = logging.getLogger(__name__)

ID = 'id'

NAME = 'name'

PART_EXCLUDE = [NAME]

PROPERTY_EXCLUDE = [NAME]

PARAMETER_EXCLUDE = [ID]

CONTROL_EXCLUDE = [ID]

CATALOG_EXCLUDE = ['uuid', 'metadata', 'back_matter']

ITEM_EXCLUDE_MAP = {
    'Part': PART_EXCLUDE,
    'Property': PROPERTY_EXCLUDE,
    'Parameter': PARAMETER_EXCLUDE,
    'Control': CONTROL_EXCLUDE,
    'Catalog': CATALOG_EXCLUDE
}


class Merge(Pipeline.Filter):
    """
    Merge the incoming catalogs according to rules in the profile.

    The incoming catalogs have already been pruned based on the import.
    Now the controls must be gathered, merged, and grouped based on the merge settings.
    """

    def __init__(self, profile: prof.Profile) -> None:
        """Initialize the class with the profile."""
        logger.debug('merge filter initialize')
        self._profile = profile

    def _get_id(self, item: OBT) -> Optional[str]:
        id_ = getattr(item, ID, None)
        if id_ is None:
            id_ = getattr(item, NAME, None)
        return id_

    def _merge_lists(self, dest: List[OBT], src: List[OBT], merge_method: Optional[prof.Method]) -> None:
        added_items = []
        if merge_method == prof.Method.keep:
            dest.extend(src)
            return
        for item in src:
            # if there is an exact copy of this in dest then ignore it
            if item not in dest:
                merged = False
                item_id = self._get_id(item)
                if item_id is not None:
                    for other in dest:
                        other_id = self._get_id(other)
                        if other_id == item_id:
                            if merge_method == prof.Method.merge:
                                self._merge_items(other, item, merge_method)
                            merged = True
                            break
                # it isn't already in dest and no match was found for merge, so append
                if not merged:
                    added_items.append(item)
        dest.extend(added_items)

    def _merge_attrs(
        self, dest: Union[OBT, List[OBT]], src: Union[OBT, List[OBT]], attr: str, merge_method: Optional[prof.Method]
    ) -> None:
        """Merge this attr of src into the attr of dest."""
        src_attr = getattr(src, attr, None)
        if src_attr is None:
            return
        item_type = type(src).__name__
        if attr in ITEM_EXCLUDE_MAP.get(item_type, []):
            return
        dest_attr = getattr(dest, attr, None)
        if dest_attr and isinstance(dest_attr, list):
            self._merge_lists(dest_attr, src_attr, merge_method)
            setattr(dest, attr, dest_attr)
            return
        if dest_attr and merge_method == prof.Method.use_first:
            return
        if dest_attr == src_attr and merge_method not in [None, prof.Method.keep]:
            return
        setattr(dest, attr, src_attr)

    def _merge_items(self, dest: OBT, src: OBT, merge_method: Optional[prof.Method]) -> None:
        """Merge two items recursively."""
        for field in src.__fields_set__:
            self._merge_attrs(dest, src, field, merge_method)

    def _group_contents(self, group: cat.Group) -> Tuple[List[cat.Control], List[com.Parameter]]:
        """Get flattened content of group and its groups recursively."""
        controls = []
        params = []
        controls.extend(as_list(group.controls))
        params.extend(as_list(group.params))
        if group.groups is not None:
            for sub_group in group.groups:
                new_controls, new_params = self._group_contents(sub_group)
                controls.extend(new_controls)
                params.extend(new_params)
        return controls, params

    def _flatten_catalog(self, catalog: cat.Catalog, as_is: bool) -> cat.Catalog:
        """Flatten the groups of the catalog if as_is is False."""
        if as_is or catalog.groups is None:
            return catalog

        # as_is is False so flatten the controls into a single list
        catalog.controls = as_list(catalog.controls)
        catalog.params = as_list(catalog.params)
        for group in catalog.groups:
            new_controls, new_params = self._group_contents(group)
            catalog.controls.extend(new_controls)
            catalog.params.extend(new_params)
        catalog.controls = none_if_empty(catalog.controls)
        catalog.params = none_if_empty(catalog.params)
        catalog.groups = None
        return catalog

    def _merge_two_catalogs(
        self, dest: cat.Catalog, src: cat.Catalog, merge_method: Optional[prof.Method], as_is: bool
    ) -> cat.Catalog:
        # merge_method is use_first, merge, keep
        # no combine or merge_method equates to merge_method=keep
        # if as_is is false, the result is flattened

        dest = self._flatten_catalog(dest, as_is)
        src = self._flatten_catalog(src, as_is)

        self._merge_items(dest, src, merge_method)

        return dest

    def _merge_catalog(self, merged: Optional[cat.Catalog], catalog: cat.Catalog) -> cat.Catalog:
        """Merge the controls in the catalog into merged catalog."""
        # no merge means keep, including dups
        # same for merge with no combine
        # groups are merged only if separate directive such as as-is is given
        # use-first is a merge combination rule
        # merge is a merge combination rule for controls.  groups are not merged by this rule
        # merge/as-is and merge/custom are used for merging groups
        # if neither as-is nor custom is specified - just get single list of controls
        # unstructured controls should appear after any loose params

        # make copies to avoid changing input objects
        local_cat = catalog.copy(deep=True)
        local_merged = merged.copy(deep=True) if merged else None

        merge_method = prof.Method.keep
        as_is = False
        if self._profile.merge is not None:
            if self._profile.merge.custom is not None:
                raise TrestleError('Profile with custom merge is not supported.')
            if self._profile.merge.as_is is not None:
                as_is = self._profile.merge.as_is
            if self._profile.merge.combine is None:
                logger.debug('Profile has merge but no combine so defaulting to combine/merge.')
                merge_method = prof.Method.merge
            else:
                merge_combine = self._profile.merge.combine
                if merge_combine.method is None:
                    logger.debug('Profile has merge combine but no method.  Defaulting to merge.')
                    merge_method = prof.Method.merge
                else:
                    merge_method = merge_combine.method

        if local_merged is None:
            return self._flatten_catalog(local_cat, as_is)

        # merge the incoming catalog with merged based on merge_method and as_is
        return self._merge_two_catalogs(local_merged, local_cat, merge_method, as_is)

    def process(self, pipelines: List[Pipeline]) -> Iterator[cat.Catalog]:  # type: ignore
        """
        Merge the incoming catalogs.

        This pulls from import and iterates over the incoming catalogs.
        The way groups, lists of controls, and controls themselves get merged is specified by the profile.
        """
        merged: Optional[cat.Catalog] = None
        logger.debug(f'merge entering process with {len(pipelines)} pipelines')
        for pipeline in pipelines:
            catalog = next(pipeline.process(None))
            merged = self._merge_catalog(merged, catalog)
        yield merged  # type: ignore



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handle direct IO for writing SSP responses as markdown."""
import logging
import pathlib
from typing import Dict, List, Optional, Tuple

from trestle.common.common_types import TypeWithByComps
from trestle.common.const import CONTROL_ORIGINATION, IMPLEMENTATION_STATUS, ITEM, SSP_MAIN_COMP_NAME
from trestle.common.const import STATUS_OPERATIONAL
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_list
from trestle.core.catalog import catalog_interface
from trestle.core.catalog.catalog_interface import CatalogInterface
from trestle.core.control_interface import ControlInterface
from trestle.core.docs_control_writer import DocsControlWriter
from trestle.core.markdown.docs_markdown_node import DocsMarkdownNode
from trestle.core.markdown.md_writer import MDWriter
from trestle.oscal import ssp
from trestle.oscal.catalog import Catalog

logger = logging.getLogger(__name__)


class SSPMarkdownWriter():
    """
    Class to write control responses as markdown.

    Functions in this class are mainly used by jinja and not by the trestle code itself.
    """

    def __init__(self, trestle_root: pathlib.Path) -> None:
        """Initialize the class."""
        self._trestle_root = trestle_root
        self._ssp: ssp.SystemSecurityPlan = None
        self._resolved_catalog: Catalog = None
        self._catalog_interface: CatalogInterface = None

    def set_ssp(self, ssp: ssp.SystemSecurityPlan) -> None:
        """Set ssp."""
        self._ssp = ssp

    def set_catalog(self, resolved_catalog: Catalog) -> None:
        """Set catalog."""
        self._resolved_catalog = resolved_catalog
        self._catalog_interface = catalog_interface.CatalogInterface(self._resolved_catalog)

    def get_control_statement(self, control_id: str, level: int) -> str:
        """
        Get the control statement for an ssp - to be printed in markdown as a structured list.

        Args:
            control_id: The control_id to use.

        Returns:
            A markdown blob as a string.
        """
        if not self._resolved_catalog:
            raise TrestleError('Cannot get control statement, set resolved catalog first.')

        writer = DocsControlWriter()
        control = self._catalog_interface.get_control(control_id)
        if not control:
            return ''

        control_lines = writer.get_control_statement_ssp(control)

        return self._build_tree_and_adjust(control_lines, level)

    def get_control_part(self, control_id: str, part_name: str, level: int) -> str:
        """Get control part with given name."""
        control_part = self._catalog_interface.get_control_part_prose(control_id, part_name)

        md_list = self._write_str_with_header(
            f'Control Part: {part_name} for control: {control_id}', control_part, level
        )
        return self._build_tree_and_adjust(md_list.split('\n'), level)

    def get_fedramp_control_tables(self, control_id: str, level: int, label_column: bool = False) -> str:
        """Get the fedramp metadata as markdown tables, with optional third label column for params.

        The fedramp metadata has the following elements:
        - Responsible roles field
        - Parameter values table
        - Implementation status field
        - Control origination field

        Returns:
            tables as one coherent markdown blob.
        """
        resp_roles_table = self.get_responsible_roles_table(control_id, level)
        params_values = self._parameter_table(control_id, level, label_column)
        impl_status = self.get_fedramp_implementation_status(control_id, level)
        control_orig = self.get_fedramp_control_origination(control_id, level)

        final_output = ''
        if resp_roles_table:
            final_output += resp_roles_table
        if params_values:
            final_output += '\n' + params_values
        if impl_status:
            final_output += '\n' + impl_status
        if control_orig:
            final_output += '\n' + control_orig
        return final_output

    def get_responsible_roles_table(self, control_id: str, level: int) -> str:
        """
        For each role id - if the role exists in metadata use the title as what gets printed in the roles table.

        If not (for now) warn and use the role-id as the printed text.
        """
        if self._ssp is None:
            raise TrestleError('Cannot get responsible roles, SSP is not set.')

        for impl_requirement in self._ssp.control_implementation.implemented_requirements:
            if impl_requirement.control_id == control_id:
                if impl_requirement.responsible_roles:
                    resp_roles = as_list(impl_requirement.responsible_roles)
                    role_ids = [role.role_id.replace('_', ' ') for role in resp_roles]

                    # now check if this role exists in the metadata
                    role_titles = dict(zip(role_ids, role_ids))
                    roles = as_list(self._ssp.metadata.roles)
                    for role in roles:
                        if role.id in role_ids:
                            role_titles[role.id] = role.title

                    # dictionary to md table
                    md_list = self._write_table_with_header(
                        'Responsible Roles.', [[key, role_titles[key]] for key in role_titles.keys()],
                        ['Role ID', 'Title'],
                        level
                    )
                    return md_list
                else:
                    logger.warning(
                        f'No responsible roles were found for the control with id: {control_id} in given SSP.'
                    )
                    return ''

        return ''

    def _parameter_table(self, control_id: str, level: int, label_column: bool = False) -> str:
        """Print Param_id | ValueOrLabelOrChoices | Optional Label Column."""
        if not self._ssp:
            raise TrestleError('Cannot get parameter table, set SSP first.')

        writer = DocsControlWriter()
        control = self._catalog_interface.get_control(control_id)
        if not control:
            return ''
        params_lines = writer.get_param_table(control, label_column)
        # need to make sure no params still have moustaches.  convert to brackets to avoid jinja complaints
        clean_lines = []
        for line in params_lines:
            clean_lines.append(line.replace('{{', '[[').replace('}}', ']]'))

        tree = DocsMarkdownNode.build_tree_from_markdown(clean_lines)
        tree.change_header_level_by(level)
        return tree.content.raw_text

    def get_fedramp_implementation_status(self, control_id: str, level: int) -> str:
        """
        Print implementation status as a list of items, only showing those that are applicable for the control.

        This is unlike the word document FedRAMP which uses checkboxes on standard set of options.
        Using a LUT to map between structured data fields, defined by FedRAMP and historical text.
        """
        if not self._ssp:
            raise TrestleError('Cannot get Fedramp implementation status, set SSP first.')

        implementation_statuses: List[str] = []
        control_impl_req = self._control_implemented_req(control_id)
        if control_impl_req and control_impl_req.props:
            for prop in control_impl_req.props:
                if prop.name == IMPLEMENTATION_STATUS:
                    implementation_statuses.append(prop.value)

        md_list = self._write_list_with_header('FedRamp Implementation Status.', implementation_statuses, level)
        return md_list

    def get_fedramp_control_origination(self, control_id: str, level: int) -> str:
        """
        Print control origination, as a list of items, only showing those that are applicable for the control.

        Using a LUT to map between structured data fields, defined by FedRAMP and historical text.
        """
        if not self._ssp:
            raise TrestleError('Cannot get FedRamp control origination, set SSP first.')

        control_origination = []
        control_impl_req = self._control_implemented_req(control_id)

        if control_impl_req and control_impl_req.props:
            for prop in control_impl_req.props:
                if prop.name == CONTROL_ORIGINATION:
                    control_origination.append(prop.value)

        md_list = self._write_list_with_header('FedRamp Control Origination.', control_origination, level)
        return md_list

    @staticmethod
    def _write_component_prompt(
        md_writer: MDWriter,
        comp_name: str,
        prose: str,
        rules: List[str],
        status: str,
        show_rules: bool,
        show_status: bool
    ) -> None:
        header = f'Component: {comp_name}'
        md_writer.new_header(1, header)
        md_writer.set_indent_level(-1)
        md_writer.new_line(prose)
        md_writer.set_indent_level(-1)
        if rules and show_rules:
            md_writer.new_header(2, title='Rules:')
            md_writer.set_indent_level(-1)
            md_writer.new_list(rules)
            md_writer.set_indent_level(-1)
        if status and show_status:
            md_writer.new_header(2, title=f'Implementation Status: {status}')

    def get_control_response(
        self,
        control_id: str,
        level: int,
        write_empty_responses: bool = False,
        show_comp: bool = True,
        show_rules: bool = False,
        show_status: bool = True
    ) -> str:
        """
        Get the full control implemented requirements, broken down based on the available control responses.

        Args:
            control_id: id of the control
            level: level of indentation
            write_empty_responses: write response even if empty
            show_comp: show the component name in the response

        Notes:
            This is intended to be invoked from a jinja template that has already written out the prompt for
            control response
        """
        if not self._resolved_catalog:
            raise TrestleError('Cannot get control response, set resolved catalog first.')

        control = self._catalog_interface.get_control(control_id)
        imp_req = self._control_implemented_req(control_id)
        if not imp_req:
            logger.info(f'No implemented requirements found for the control {control_id}')
            return ''

        md_writer = MDWriter(None)

        system_prose = ''
        system_rules = []
        system_status = STATUS_OPERATIONAL
        imp_req_responses = self._get_responses_by_components(imp_req, write_empty_responses)
        if SSP_MAIN_COMP_NAME in imp_req_responses:
            system_prose, system_rules, system_status = imp_req_responses[SSP_MAIN_COMP_NAME]

        SSPMarkdownWriter._write_component_prompt(
            md_writer, SSP_MAIN_COMP_NAME, system_prose, system_rules, system_status, show_rules, show_status
        )

        # if a control has no statement sub-parts then get the response bycomps from the imp_req itself
        # otherwise get them from the statements in the imp_req
        # an imp_req and a statement are both things that can have bycomps
        has_bycomps = imp_req.statements if imp_req.statements else [imp_req]
        for has_bycomp in has_bycomps:
            statement_id = getattr(has_bycomp, 'statement_id', f'{control_id}_smt')
            label = statement_id
            part_name = None

            # look up label for this statement
            if control.parts:
                found_label, part = self._catalog_interface.get_statement_label_if_exists(control_id, statement_id)
                if found_label:
                    label = found_label
                    part_name = part.name

            response_per_component = self._get_responses_by_components(has_bycomp, write_empty_responses)

            if response_per_component or write_empty_responses:
                if part_name and part_name == ITEM:
                    # print part header only if subitem
                    header = f'Implementation for part {label}'
                    md_writer.new_header(1, title=header)
                for comp_name, comp_response in response_per_component.items():
                    if comp_name == SSP_MAIN_COMP_NAME:
                        continue
                    prose, rules, status = comp_response
                    if show_comp:
                        SSPMarkdownWriter._write_component_prompt(
                            md_writer, comp_name, prose, rules, status, show_rules, show_status
                        )

        lines = md_writer.get_lines()

        tree = DocsMarkdownNode.build_tree_from_markdown(lines)
        tree.change_header_level_by(level)

        return tree.content.raw_text

    def _get_responses_by_components(self, has_bycomps: TypeWithByComps,
                                     write_empty_responses: bool) -> Dict[str, Tuple[str, List[str], str]]:
        """Get response per component, substitute component id with title if possible."""
        response_per_component: Dict[str, Tuple[str, str]] = {}
        for by_comp in as_list(has_bycomps.by_components):  # type: ignore
            # look up component title
            subheader = by_comp.component_uuid
            prose = ''
            status = ''
            rules = []
            if self._ssp.system_implementation.components:
                for comp in self._ssp.system_implementation.components:
                    if comp.uuid == by_comp.component_uuid:
                        title = comp.title
                        if title:
                            subheader = title
            if by_comp.description:
                prose = by_comp.description
            if by_comp.implementation_status:
                status = by_comp.implementation_status.state
            rules, _ = ControlInterface.get_rule_list_for_item(by_comp)

            if prose or (not prose and write_empty_responses):
                if subheader:
                    response_per_component[subheader] = (prose, rules, status)

        return response_per_component

    def _control_implemented_req(self, control_id: str) -> Optional[ssp.ImplementedRequirement]:
        """Retrieve control implemented requirement by control-id."""
        requirements = self._ssp.control_implementation.implemented_requirements
        for requirement in requirements:
            if requirement.control_id == control_id:
                return requirement
        logger.debug(f'No implemented requirement found for control {control_id}')
        return None

    def _write_list_with_header(self, header: str, lines: List[str], level: int) -> str:
        if lines:
            md_writer = MDWriter(None)
            md_writer.new_paragraph()
            md_writer.new_header(level=1, title=header)
            md_writer.set_indent_level(-1)
            md_writer.new_list(lines)
            md_writer.set_indent_level(-1)

            return self._build_tree_and_adjust(md_writer.get_lines(), level)

        return ''

    def _write_table_with_header(
        self, header: str, values: List[List[str]], table_header: List[str], level: int
    ) -> str:
        if values and values[0]:
            md_writer = MDWriter(None)
            md_writer.new_paragraph()
            md_writer.new_header(level=1, title=header)
            md_writer.set_indent_level(-1)
            md_writer.new_table(values, table_header)
            md_writer.set_indent_level(-1)

            return self._build_tree_and_adjust(md_writer.get_lines(), level)
        return ''

    def _write_str_with_header(self, header: str, text: str, level: int) -> str:
        if text:
            md_writer = MDWriter(None)
            md_writer.new_paragraph()
            md_writer.new_header(level=1, title=header)
            md_writer.set_indent_level(-1)
            md_writer.new_line(text)
            md_writer.set_indent_level(-1)

            return self._build_tree_and_adjust(md_writer.get_lines(), level)
        return ''

    def _build_tree_and_adjust(self, lines: List[str], level: int) -> str:
        tree = DocsMarkdownNode.build_tree_from_markdown(lines)
        tree.change_header_level_by(level)

        return tree.content.raw_text



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for md_writer module."""

import pathlib

from tests.test_utils import confirm_text_in_file

from trestle.core.markdown.md_writer import MDWriter


def test_md_writer(tmp_path: pathlib.Path) -> None:
    """Test md_writer."""
    md_file = tmp_path / 'md_file.md'
    md_writer = MDWriter(md_file)
    md_writer.set_indent_step_size(2)
    md_writer.new_paragraph()
    md_writer.new_header(level=2, title='Control Statement')
    md_writer.set_indent_level(-2)
    items = [
        'The organization', ['a thing', ['1. thing', '2. thing']], ['b thing'], ['c thing', ['1. things', '2. things']],
        ['d thing']
    ]
    md_writer.new_list(items)
    md_writer.new_paragraph()
    md_writer.new_hr()
    md_writer.new_line('my line')
    header = {'a': 1, 'b': {'x': 2, 'y': 3}, 'c': 4}
    md_writer.add_yaml_header(header)
    md_writer.new_paragraph()
    md_writer.write_out()

    desired_result = """---
a: 1
b:
  x: 2
  y: 3
c: 4
---

## Control Statement

- The organization

- a thing

  - 1. thing
  - 2. thing

- b thing

- c thing

  - 1. things
  - 2. things

- d thing

______________________________________________________________________

my line
"""

    with open(md_file) as f:
        md_result = f.read()
    assert desired_result == md_result


def test_cull_headings(testdata_dir: pathlib.Path, tmp_path: pathlib.Path) -> None:
    """Test culling of headings from md tree."""
    markdown_file = testdata_dir / 'markdown/valid_complex_md.md'
    strict_cull_list = ['## 1.2 MD Subheader 1.2', '### 1.3.1 Valid header <!-- ### some comment here -->']
    # make sure the headers are present in the original
    for item in strict_cull_list:
        assert confirm_text_in_file(markdown_file, '', item)
    strict_path = tmp_path / 'strict.md'
    md_writer = MDWriter(strict_path)
    md_writer.cull_headings(markdown_file, strict_cull_list, True)
    # make sure headers are gone now
    for item in strict_cull_list:
        assert not confirm_text_in_file(strict_path, '', item)
    non_strict_path = tmp_path / 'non_strict.md'
    md_writer = MDWriter(non_strict_path)
    non_strict_cull_list = ['Subheader', 'Header']
    md_writer.cull_headings(markdown_file, non_strict_cull_list, False)
    for item in strict_cull_list:
        assert not confirm_text_in_file(non_strict_path, '', item)
    ignore_path = tmp_path / 'ignore.md'
    md_writer = MDWriter(ignore_path)
    ignore_list = ['IgnoreIt']
    md_writer.cull_headings(markdown_file, ignore_list, False)
    for letter in 'ABCDEFG':
        assert confirm_text_in_file(ignore_path, '', f'IgnoreIt {letter}')
